from .d64_bam import D64BAM


class D71BAM(D64BAM):

    FIRST_TRACK_ON_REVERSE = 36

    def get_entry(self, track):
        """Return a tuple of total free blocks and a string of free blocks for a track."""
        if track < self.FIRST_TRACK_ON_REVERSE:
            # tracks below 36 are as for a d64 image
            return super().get_entry(track)
        if track > self.image.MAX_TRACK:
            raise ValueError("Invalid track, %d" % track)

        # tracks 36 and above are split between 18/0 and 53/0
        total = self.image.bam_block.get(0xb9+track)
        start = (track-self.FIRST_TRACK_ON_REVERSE)*(self.BAM_ENTRY_SIZE-1)
        e = tuple(self.image.extra_bam_block.get(start, start+self.BAM_ENTRY_SIZE-1))
        free_bits = ''
        for b in e:
            free_bits += ''.join(reversed(format(b, '08b')))
        return total, free_bits

    def set_entry(self, track, total, free_bits):
        """Update the block allocation entry for a track."""
        if track < self.FIRST_TRACK_ON_REVERSE:
            # tracks below 36 are as for a d64 image
            super().set_entry(track, total, free_bits)
            return
        if track > self.image.MAX_TRACK:
            raise ValueError("Invalid track, %d" % track)

        # tracks 36 and above are split between 18/0 and 53/0
        self.image.bam_block.set(0xb9+track, total)
        start = (track-self.FIRST_TRACK_ON_REVERSE)*(self.BAM_ENTRY_SIZE-1)
        entry = []
        while free_bits:
            val = ''.join(reversed(free_bits[:8]))
            entry.append(int(val, 2))
            free_bits = free_bits[8:]
        bin_entry = bytes(entry)
        self.image.extra_bam_block.set(start, bin_entry)



from .block import Block
from .d64_image import D64Image
from .d71_bam import D71BAM


class D71Image(D64Image):

    DOS_VERSION = ord('2')
    DOS_FORMAT = ord('A')
    MAX_TRACK = 70
    EXTRA_BAM_TRACK = 53
    INTERLEAVE = 6
    DIR_INTERLEAVE = 3
    TRACK_SECTOR_MAX = ((21, (1, 17)), (19, (18, 24)), (18, (25, 30)), (17, (31, 35)),
                        (21, (36, 52)), (19, (53, 59)), (18, (60, 65)), (17, (66, 70)))
    IMAGE_SIZES = (349696, 351062)

    def __init__(self, filename):
        super().__init__(filename)
        self.bam = D71BAM(self)

    def open(self, mode):
        super().open(mode)
        self.extra_bam_block = Block(self, self.EXTRA_BAM_TRACK, 0)

    def alloc_next_block(self, track, sector, directory=False):
        return self._alloc_next_block(track, sector, self.DIR_INTERLEAVE if directory else self.INTERLEAVE)

    @classmethod
    def create(cls, filepath, disk_name, disk_id):
        """Create an empty disk image."""
        super().create(filepath, disk_name, disk_id)

        image = cls(filepath)
        try:
            image.open('r+b')

            # block 18/0 contains part of the BAM and various identifying fields
            bam_block = Block(image, cls.DIR_TRACK, 0)
            bam_block.set(0x90, b'\xa0' * 0x1b)
            image.name = disk_name
            image.id = disk_id
            image.dos_type = cls.DOS_FORMAT
            image.dos_version = cls.DOS_VERSION

            # populate the BAM with all free blocks
            for sectors, range_ in cls.TRACK_SECTOR_MAX:
                for t in range(range_[0], range_[1]+1):
                    bits_free = '1' * sectors
                    image.bam.set_entry(t, sectors, bits_free)

            # block 18/1 contains 8 empty directory entries
            dir_block = Block(image, cls.DIR_TRACK, cls.DIR_SECTOR)
            dir_block.data_size = 0xfe

            # link BAM block to directory block
            bam_block.set_next_block(dir_block)

            # allocate both blocks
            image.bam.set_allocated(cls.DIR_TRACK, 0)
            image.bam.set_allocated(cls.DIR_TRACK, 1)

            # all of track 53 is allocated
            _, free_bits = image.bam.get_entry(cls.EXTRA_BAM_TRACK)
            free_bits = free_bits.replace('1', '0')
            image.bam.set_entry(cls.EXTRA_BAM_TRACK, 0, free_bits)
        finally:
            image.close()



from .block import Block
from .d81_relative_file import D81RelativeFile
from .dir_entry import DirEntry


class D81DirEntry(DirEntry):
    FTYPE_STR = ('DEL', 'SEQ', 'PRG', 'USR', 'REL', 'CBM', '???', '???')

    def free_blocks(self):
        """Free any data blocks and side sectors."""
        if self.file_type == 'CBM':
            for block in self.partition_blocks():
                block.image.free_block(block)
        else:
            super().free_blocks()

    def relative_file(self, mode):
        """Return relative file instance."""
        return D81RelativeFile(self, mode)

    def partition_blocks(self):
        """Generator to return each block of a partition."""
        block = self.first_block()
        block_count = self.size

        while block_count:
            yield block

            block_count -= 1
            if block_count:
                # blocks are allocated contiguously
                sector = block.sector+1
                if sector >= block.image.max_sectors(block.track):
                    # end of current track
                    track = block.track+1
                    sector = 0
                else:
                    track = block.track
                block = Block(block.image, track, sector)



from .block import Block
from .d64_bam import D64BAM
from .dos_image import DOSImage


class D64Image(DOSImage):

    DOS_VERSION = ord('2')
    DOS_FORMAT = ord('A')
    MAX_TRACK = 35
    DIR_TRACK = 18
    DIR_SECTOR = 1
    INTERLEAVE = 10
    DIR_INTERLEAVE = 3
    TRACK_SECTOR_MAX = ((21, (1, 17)), (19, (18, 24)), (18, (25, 30)), (17, (31, 35)))
    IMAGE_SIZES = (174848, 175531)

    def __init__(self, filename):
        super().__init__(filename)
        self.bam = D64BAM(self)

    def open(self, mode):
        super().open(mode)
        self.bam_block = Block(self, self.DIR_TRACK, 0)

    def alloc_next_block(self, track, sector, directory=False):
        return self._alloc_next_block(track, sector, self.DIR_INTERLEAVE if directory else self.INTERLEAVE)

    @property
    def dos_version(self):
        return self.bam_block.get(0xa5)

    @property
    def name(self):
        name = self.bam_block.get(0x90, 0xa0)
        return name.rstrip(b'\xa0')

    @property
    def id(self):
        id = self.bam_block.get(0xa2, 0xa4)
        return id

    @property
    def dos_type(self):
        return self.bam_block.get(2)

    @dos_version.setter
    def dos_version(self, version):
        self.bam_block.set(0xa5, version)

    @name.setter
    def name(self, nam):
        self.bam_block.set(0x90, nam[:16].ljust(16, b'\xa0'))

    @id.setter
    def id(self, did):
        if len(did) != 2:
            raise ValueError("Invalid disk id, "+str(did))
        self.bam_block.set(0xa2, did)

    @dos_type.setter
    def dos_type(self, dtype):
        self.bam_block.set(2, dtype)
        self.bam_block.set(0xa6, dtype)

    @classmethod
    def create(cls, filepath, disk_name, disk_id):
        """Create an empty disk image."""
        super().create(filepath)

        image = cls(filepath)
        try:
            image.open('r+b')

            # block 18/0 contains the BAM and various identifying fields
            bam_block = Block(image, cls.DIR_TRACK, 0)
            bam_block.set(0x90, b'\xa0' * 0x1b)
            image.name = disk_name
            image.id = disk_id
            image.dos_type = cls.DOS_FORMAT
            image.dos_version = cls.DOS_VERSION

            # populate the BAM with all free blocks
            for sectors, range_ in cls.TRACK_SECTOR_MAX:
                for t in range(range_[0], range_[1]+1):
                    bits_free = '1' * sectors
                    image.bam.set_entry(t, sectors, bits_free)

            # block 18/1 contains 8 empty directory entries
            dir_block = Block(image, cls.DIR_TRACK, cls.DIR_SECTOR)
            dir_block.data_size = 0xfe

            # link BAM block to directory block
            bam_block.set_next_block(dir_block)

            # allocate both blocks
            image.bam.set_allocated(cls.DIR_TRACK, 0)
            image.bam.set_allocated(cls.DIR_TRACK, 1)
        finally:
            image.close()



import logging
import sys

from d64 import Block, DiskImage


logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)

part_name = sys.argv[2].encode()
subdir_id = sys.argv[3].encode()

part_start = 1
part_size = 120

with DiskImage(sys.argv[1], 'w') as i:
    p = i.partition(part_name)
    p.create(Block(i, part_start, 0), part_size)
    print(p.entry, p.entry.block)
    print(p.entry.file_type, p.name)
    p.format(part_name, subdir_id)

    s = i.subdirectory(part_name)
    print(s.id)



import sys

from d64 import Block, DiskImage

part_name = sys.argv[2].encode()
with DiskImage(sys.argv[1], 'w') as i:
    p = i.partition(part_name)
    p.create(Block(i, 1, 0), 10)
    print(p.entry, p.entry.block)
    print(p.entry.file_type, p.name)



import unittest

from d64.d82_image import D82Image


class TestD82ImageBlocks(unittest.TestCase):

    def setUp(self):
        self.image = D82Image(None)

    def test_max_sectors(self):
        self.assertEqual(self.image.max_sectors(1), 29)
        self.assertEqual(self.image.max_sectors(50), 27)
        self.assertEqual(self.image.max_sectors(60), 25)
        self.assertEqual(self.image.max_sectors(77), 23)
        self.assertEqual(self.image.max_sectors(134), 25)
        self.assertEqual(self.image.max_sectors(154), 23)

    def test_max_sectors_bad(self):
        with self.assertRaises(ValueError):
            self.image.max_sectors(0)
        with self.assertRaises(ValueError):
            self.image.max_sectors(155)

    def test_block_start(self):
        self.assertEqual(self.image.block_start(1, 0), 0)
        self.assertEqual(self.image.block_start(1, 1), 256)
        self.assertEqual(self.image.block_start(78, 0), 533248)
        self.assertEqual(self.image.block_start(139, 12), 973824)

    def test_block_start_bad(self):
        with self.assertRaises(ValueError):
            self.image.block_start(0, 0)
        with self.assertRaises(ValueError):
            self.image.block_start(123, 31)


if __name__ == '__main__':
    unittest.main()



import unittest

from unittest.mock import Mock

from d64.subdirectory import Subdirectory

from test.mock_block import MockBlock


class TestSubdirectory(unittest.TestCase):

    def test_subdirectory(self):
        mock_image = Mock()
        mock_entry = Mock()
        mock_entry.size = 120
        start_block = MockBlock(mock_image, 10, 0)
        mock_entry.first_block.return_value = start_block
        subdir = Subdirectory(mock_image, mock_entry)
        self.assertEqual(subdir.MIN_TRACK, 10)
        self.assertEqual(subdir.DIR_TRACK, 10)
        self.assertEqual(subdir.MAX_TRACK, 12)



import argparse
import sys

from pathlib import Path

from d64 import DiskImage


def main():
    parser = argparse.ArgumentParser(description='Create empty Commodore disk images.')
    parser.add_argument('label', help='disk label')
    parser.add_argument('id', help='disk identifier')
    parser.add_argument('filename', type=Path, help='image filename')
    parser.add_argument('--type', default='d64', choices=DiskImage.supported_types(), help='image type')
    parser.add_argument('--force', action='store_true', help='overwrite existing image')
    args = parser.parse_args()

    if args.filename.exists():
        if args.force:
            args.filename.unlink()
        else:
            sys.exit("{!s} already exists".format(args.filename))
    print("Creating empty disk image as {!s}, {}:{}, type {}".format(args.filename, args.label, args.id, args.type))

    DiskImage.create(args.type, args.filename, args.label.encode(), args.id.encode())



import unittest
from unittest.mock import patch, Mock

from d64.dir_entry import DirEntry

from test.mock_block import MockBlock


class TestDirEntryRead(unittest.TestCase):

    def setUp(self):
        mock_image = Mock()
        mock_image.block_start.return_value = 0
        self.block = MockBlock(mock_image)
        self.block.data = bytearray(32)
        self.entry = DirEntry(self.block, 0)

    def test_read_file_type(self):
        self.block.data[2] = 0x80
        self.assertEqual(self.entry.file_type, 'DEL')
        self.assertFalse(self.entry.protected)
        self.assertFalse(self.entry.replacement)
        self.assertTrue(self.entry.closed)
        self.block.data[2] = 0x81
        self.assertEqual(self.entry.file_type, 'SEQ')
        self.block.data[2] = 0x82
        self.assertEqual(self.entry.file_type, 'PRG')
        self.block.data[2] = 0x03
        self.assertEqual(self.entry.file_type, 'USR')
        self.assertFalse(self.entry.closed)
        self.block.data[2] = 0xc4
        self.assertEqual(self.entry.file_type, 'REL')
        self.assertTrue(self.entry.protected)
        self.block.data[2] = 0x85
        self.assertEqual(self.entry.file_type, '???')
        self.block.data[2] = 0x22
        self.assertEqual(self.entry.file_type, 'PRG')
        self.assertTrue(self.entry.replacement)

    def test_read_start_ts(self):
        self.block.data[3:4] = b'\x0a\x14'
        self.assertEqual(self.entry.start_ts, (10, 20))
        first_block = self.entry.first_block()
        self.assertEqual(first_block.track, 10)
        self.assertEqual(first_block.sector, 20)

    def test_read_replacement_ts(self):
        self.block.data[0x1c:0x1e] = b'\x11\x07'
        self.assertEqual(self.entry.replacement_ts, (17, 7))

    def test_read_name(self):
        self.block.data[5:0x16] = b'\x46\x49\x47\x48\x54\x45\x52\x20\x52\x41\x49\x44\xA0\xA0\xA0\xA0'
        self.assertEqual(self.entry.name, b'FIGHTER RAID')

    def test_read_ss_ts(self):
        self.block.data[0x15:0x17] = b'\x0b\x0c'
        self.assertEqual(self.entry.side_sector_ts, (11, 12))

    def test_read_rec_len(self):
        self.block.data[0x17] = 0x50
        self.assertEqual(self.entry.record_len, 80)

    def test_read_size(self):
        self.block.data[0x1e:0x20] = b'\xe7\x01'
        self.assertEqual(self.entry.size, 487)


class TestDirEntryWrite(unittest.TestCase):

    def setUp(self):
        self.block = MockBlock()
        self.block.data = bytearray(32)
        self.entry = DirEntry(self.block, 0)

    def test_write_file_type(self):
        self.block.data[2] = 0x80
        self.entry.file_type = 'prg'
        self.assertEqual(self.block.data[2], 0x82)
        self.entry.file_type = 0xc3
        self.assertEqual(self.block.data[2], 0xc3)

    def test_write_set_first_block(self):
        self.entry.start_ts = (21, 11)
        mock_block = MockBlock(None, 32, 7)
        self.entry.set_first_block(mock_block)
        self.assertEqual(self.block.data[3:5], b'\x20\x07')

    def test_write_reset_entry(self):
        self.entry.size = 14
        self.start_ts = (7, 16)
        self.side_sector_ts = (31, 6)
        self.entry.reset()
        self.assertEqual(self.entry.size, 0)
        self.assertEqual(self.entry.start_ts, (0, 0))
        self.assertEqual(self.entry.side_sector_ts, (0, 0))

    def test_write_protected(self):
        self.block.data[2] = 0x82
        self.entry.protected = True
        self.assertEqual(self.block.data[2], 0xc2)

    def test_write_closed(self):
        self.block.data[2] = 0xc2
        self.entry.closed = False
        self.assertEqual(self.block.data[2], 0x42)

    def test_write_replacement(self):
        self.block.data[2] = 0x22
        self.entry.replacement = False
        self.assertEqual(self.block.data[2], 0x02)

    def test_write_start_ts(self):
        self.entry.start_ts = (21, 11)
        self.assertEqual(self.block.data[3:5], b'\x15\x0b')

    def test_write_replacement_ts(self):
        self.entry.replacement_ts = (8, 14)
        self.assertEqual(self.block.data[0x1c:0x1e], b'\x08\x0e')

    def test_write_name(self):
        self.block.data[5:21] = b'\xa0' * 16
        self.entry.name = b'LONG NAME'
        self.assertEqual(self.block.data[5:21], b'LONG NAME\xa0\xa0\xa0\xa0\xa0\xa0\xa0')
        self.entry.name = b'SHORT'
        self.assertEqual(self.block.data[5:21], b'SHORT\xa0\xa0\xa0\xa0\xa0\xa0\xa0\xa0\xa0\xa0\xa0')
        self.block.data[21] = 0x99
        self.entry.name = b'VERY VERY VERY LONG NAME'
        self.assertEqual(self.block.data[5:22], b'VERY VERY VERY L\x99')

    def test_write_ss_ts(self):
        self.entry.side_sector_ts = (11, 12)
        self.assertEqual(self.block.data[0x15:0x17], b'\x0b\x0c')

    def test_write_rec_len(self):
        self.entry.record_len = 130
        self.assertEqual(self.block.data[0x17], 0x82)

    def test_write_size(self):
        self.entry.size = 1234
        self.assertEqual(self.block.data[0x1e:0x20], b'\xd2\x04')


class TestDirEntryFree(unittest.TestCase):

    def setUp(self):
        self.image = Mock()
        self.image.block_start.return_value = 0
        self.block = MockBlock(self.image)
        self.block.data = bytearray(32)
        self.entry = DirEntry(self.block, 0)

    def test_free_blocks(self):
        self.entry.start_ts = (21, 11)
        with patch('d64.dir_entry.Block', new=MockBlock):
            self.entry.free_blocks()
        self.image.free_block.assert_called_once()

    def test_free_blocks_ss(self):
        freed = []

        def _free_block(b):
            freed.append((b.track, b.sector))

        self.image.free_block = _free_block
        self.entry.file_type = 'REL'
        self.entry.start_ts = (21, 11)
        self.entry.side_sector_ts = (11, 12)
        with patch('d64.dir_entry.Block', new=MockBlock):
            self.entry.free_blocks()
        self.assertIn((21, 11), freed)
        self.assertIn((11, 12), freed)



import unittest

from contextlib import suppress
from pathlib import Path

import d64.scripts.d64_fsck

from d64.d81_image import D81Image

import binary


class TestD81_fsck(unittest.TestCase):

    def setUp(self):
        self.base_path = Path(__file__).parent / 'data' / 'test.d81'
        self.test_path = Path('/tmp/test_bad.d81')
        self.base_bin = binary.load_binary(self.base_path)
        d64.scripts.d64_fsck.QUIET = True
        d64.scripts.d64_fsck.FIX = True
        d64.scripts.d64_fsck.YES = True

    def tearDown(self):
        with suppress(FileNotFoundError):
            self.test_path.unlink()

    def test_clean(self):
        d64.scripts.d64_fsck.FIX = False
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.base_path), 0)

    def test_bad_dir_link(self):
        patch = [{'at': 399361, 'from': b'\x03', 'to': b'\x01'}]
        binary.patch(self.base_bin, patch, self.test_path)
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.test_path), 0)
        image = D81Image(self.test_path)
        try:
            image.open('rb')
            self.assertEqual(image.header_block.next_block().sector, 3)
        finally:
            image.close()

    def test_bad_bam_size(self):
        patch = [{'at': 399873, 'from': b'\xff', 'to': b'\x88'}]
        binary.patch(self.base_bin, patch, self.test_path)
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.test_path), 0)
        image = D81Image(self.test_path)
        try:
            image.open('rb')
            self.assertEqual(image.side_b_bam_block.data_size, 0xfe)
        finally:
            image.close()

    def test_bam_40_02_not_alloc(self):
        patch = [{'at': 399866, 'from': b'$\xf0', 'to': b'%\xf4'}]
        binary.patch(self.base_bin, patch, self.test_path)
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.test_path), 0)
        image = D81Image(self.test_path)
        try:
            image.open('rb')
            self.assertEqual(image.bam.get_entry(40), (36, '0000111111111111111111111111111111111111'))
        finally:
            image.close()

    def test_partition_not_alloc(self):
        patch = [{'at': 399632, 'from': b'(\xff\xff\xff\xff\xff'*29, 'to': b'\x00'*168+b'\x01\x04\x00\x00\x00\x00'},
                 {'at': 400130, 'from': b'\x00\x00', 'to': b'\x85\x01'},
                 {'at': 400133, 'from': b'\x00'*16, 'to': b'PARTITION\xa0\xa0\xa0\xa0\xa0\xa0\xa0'},
                 {'at': 400158, 'from': b'\x00\x00', 'to': b'\x88\x04'}]
        binary.patch(self.base_bin, patch, self.test_path)
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.test_path), 0)
        image = D81Image(self.test_path)
        try:
            image.open('rb')
            self.assertEqual(image.bam.get_entry(29), (0, '0000000000000000000000000000000000000000'))
        finally:
            image.close()


if __name__ == '__main__':
    unittest.main()



import codecs
import logging
import mmap

import petscii_codecs  # noqa: F401

from .block import Block
from .dir_entry import DirEntry
from .dos_path import DOSPath
from .exceptions import DiskFullError

log = logging.getLogger(__name__)


class DOSImage(object):

    DIR_ENTRY = DirEntry
    MIN_TRACK = 1

    @classmethod
    def create(cls, filepath):
        """Create an empty disk image."""
        empty = bytes(cls.IMAGE_SIZES[0])
        with filepath.open('wb') as fh:
            fh.write(empty)

    @classmethod
    def valid_image(cls, filepath):
        """Return `True` if a file is an image supported by a derived class."""
        return filepath.stat().st_size in cls.IMAGE_SIZES

    def __init__(self, filename):
        self.filename = filename
        self.map = None
        self.fileh = None

    def directory(self, pattern=b'*', encoding='petscii-c64en-uc', drive=0):
        """Display entries in image."""
        try:
            _ = codecs.lookup(encoding)
        except LookupError:
            log.warning("PETSCII codecs not available, using ASCII")
            encoding = 'ascii'

        dos_info = bytes((self.dos_version, self.dos_type))
        yield '{} "{:16}" {} {}'.format(drive, self.name.decode(encoding), self.id.decode(encoding), dos_info.decode(encoding))
        for path in self.glob(pattern):
            closed_ch = ' ' if path.entry.closed else '*'
            file_type = path.entry.file_type
            if path.entry.protected:
                file_type += '<'
            yield '{:5}{:18}{}{}'.format(str(path.size_blocks), '"'+path.name.decode(encoding)+'"', closed_ch, file_type)
        text = b'BLOCKS FREE.'.decode(encoding)
        yield '{} {}'.format(self.bam.total_free(), text)

    def open(self, mode):
        self.fileh = open(self.filename, mode)
        self.writeable = mode == 'r+b'
        access = mmap.ACCESS_WRITE if self.writeable else mmap.ACCESS_READ
        self.map = mmap.mmap(self.fileh.fileno(), 0, access=access)
        self.dir_block = Block(self, self.DIR_TRACK, self.DIR_SECTOR)

    def close(self):
        if self.map:
            self.map.close()
        if self.fileh:
            self.fileh.close()

    def iterdir(self, include_deleted=False):
        """Iterate over a directory."""
        block = self.dir_block

        while block:
            for offset in range(0, Block.SECTOR_SIZE, DirEntry.ENTRY_SIZE):
                entry = self.DIR_ENTRY(block, offset)
                if include_deleted or not entry.is_deleted:
                    yield DOSPath(self, entry=entry)

            block = block.next_block()

    def glob(self, pattern, include_deleted=False):
        """Return paths that match a wildcard pattern."""
        for path in self.iterdir(include_deleted):
            if DOSPath.wildcard_match(path.entry.name, path.entry.file_type, pattern):
                yield path

    def get_free_entry(self):
        """Return an unused entry for a new path."""
        block = Block(self, self.DIR_TRACK, self.DIR_SECTOR)
        entry = None

        # search directory for first free entry
        while block:
            for offset in range(0, Block.SECTOR_SIZE, DirEntry.ENTRY_SIZE):
                e = self.DIR_ENTRY(block, offset)
                if e.is_deleted:
                    entry = e
                    break

            last_block = block
            block = block.next_block()

        if entry is None:
            # no free entry, append a new block to the directory
            block = self.alloc_next_block(last_block.track, last_block.sector, directory=True)
            if block is None:
                return None

            block.data_size = Block.SECTOR_SIZE-2
            last_block.set_next_block(block)
            entry = self.DIR_ENTRY(block, 0)

        return entry

    def path(self, name):
        """Return a DOSPath instance for an entry."""
        paths = [e for e in self.glob(name)]
        if paths:
            # existing path
            return paths[0]
        return DOSPath(self, name=name)

    def alloc_first_block(self):
        """Allocate the first block for a new file."""
        track = None
        low_range = list(range(self.DIR_TRACK-1, self.MIN_TRACK-1, -1))
        high_range = list(range(self.DIR_TRACK+1, self.MAX_TRACK+1))
        range_len = max(len(low_range), len(high_range))
        low_range += [None]*range_len
        high_range += [None]*range_len
        for low, high in zip(low_range[:range_len], high_range[:range_len]):
            if low is not None:
                total, free_bits = self.bam.get_entry(low)
                if total:
                    track = low
                    break
            if high is not None:
                total, free_bits = self.bam.get_entry(high)
                if total:
                    track = high
                    break

        if track is None:
            # tried either side of the directory, disk is full
            return None
        sector = self.bam.free_from(free_bits, 0)

        self.bam.set_allocated(track, sector)
        return Block(self, track, sector)

    def _alloc_next_block(self, track, sector, interleave):
        """Allocate a subsequent block for a file."""
        cur_track = track
        cur_sector = sector
        delta = -1 if track < self.DIR_TRACK else 1
        tries = 2  # either side of directory track

        while True:
            total, free_bits = self.bam.get_entry(cur_track)
            if total:
                # free sector in current track
                cur_sector += interleave
                if cur_sector >= self.max_sectors(cur_track):
                    cur_sector -= self.max_sectors(cur_track)
                    if cur_sector:
                        cur_sector -= 1

                cur_sector = self.bam.free_from(free_bits, cur_sector)
                self.bam.set_allocated(cur_track, cur_sector)
                return Block(self, cur_track, cur_sector)

            if cur_track == self.DIR_TRACK:
                # tried either side of the directory then the directory, disk is full
                return None

            cur_track += delta
            if cur_track < self.MIN_TRACK or cur_track > self.MAX_TRACK:
                # end of disk
                cur_sector = 0
                tries -= 1
                if tries:
                    # try other side of directory
                    delta = -delta
                    cur_track = self.DIR_TRACK + delta
                else:
                    # tried either side of the directory, try directory track
                    cur_track = self.DIR_TRACK

    def free_block(self, block):
        """Mark a block as free in the BAM."""
        self.bam.set_free(block.track, block.sector)

    def block_start(self, track, sector):
        """Return offset into image of a block."""
        if track < self.MIN_TRACK or track > self.MAX_TRACK:
            raise ValueError("Invalid track, %d" % track)

        sector_start = 0
        for sectors, track_range in self.TRACK_SECTOR_MAX:
            if track > track_range[1]:
                sector_start += (track_range[1]-track_range[0]+1) * sectors
            else:
                if sector >= sectors:
                    raise ValueError("Invalid sector, %d:%d" % (track, sector))
                sector_start += (track-track_range[0]) * sectors
                sector_start += sector
                break

        return sector_start*Block.SECTOR_SIZE

    def max_sectors(self, track):
        """Return the maximum sectors for a given track."""
        if track < self.MIN_TRACK:
            raise ValueError("Invalid track, %d" % track)

        for sectors, track_range in self.TRACK_SECTOR_MAX:
            if track <= track_range[1]:
                return sectors

        raise ValueError("Invalid track, %d" % track)

    def clone_chain(self, block):
        """Return a duplicate of a chain of blocks."""
        new_block = self.alloc_next_block(block.track, block.sector)
        if new_block is None:
            raise DiskFullError

        new_block.set(0, block.get(0, block.SECTOR_SIZE))

        if not block.is_final:
            # recurse for next block
            new_block.set_next_block(self.clone_chain(block.next_block()))

        return new_block

    def __str__(self):
        return "{}({})".format(type(self).__name__, self.filename)



from .bam import BAM


class D64BAM(BAM):

    BAM_OFFSET = 4
    BAM_ENTRY_SIZE = 4

    def get_entry(self, track):
        """Return a tuple of total free blocks and a string of free blocks for a track."""
        if track < self.image.MIN_TRACK or track > self.image.MAX_TRACK:
            raise ValueError("Invalid track, %d" % track)

        start = self.BAM_OFFSET+(track-1)*self.BAM_ENTRY_SIZE
        e = tuple(self.image.bam_block.get(start, start+self.BAM_ENTRY_SIZE))
        free_bits = ''
        for b in e[1:]:
            free_bits += ''.join(reversed(format(b, '08b')))

        return e[0], free_bits

    def set_entry(self, track, total, free_bits):
        """Update the block allocation entry for a track."""
        if track < self.image.MIN_TRACK or track > self.image.MAX_TRACK:
            raise ValueError("Invalid track, %d" % track)

        start = self.BAM_OFFSET+(track-1)*self.BAM_ENTRY_SIZE
        entry = [total]
        while free_bits:
            val = ''.join(reversed(free_bits[:8]))
            entry.append(int(val, 2))
            free_bits = free_bits[8:]
        bin_entry = bytes(entry)
        self.image.bam_block.set(start, bin_entry)



import unittest

from d64.d64_image import D64Image

from test.mock_block import MockBlock


class TestD64ImageRead(unittest.TestCase):

    def setUp(self):
        self.image = D64Image(None)
        self.image.bam_block = MockBlock()

    def test_read_dos_version(self):
        self.image.bam_block.data[0xa5] = 0x32
        self.assertEqual(self.image.dos_version, ord('2'))

    def test_read_name(self):
        self.image.bam_block.data[0x90:0xa0] = b'GAMES TAPE\xa0\xa0\xa0\xa0\xa0\xa0\xa0\xa0'
        self.assertEqual(self.image.name, b'GAMES TAPE')

    def test_read_id(self):
        self.image.bam_block.data[0xa2:0xa4] = b'GT'
        self.assertEqual(self.image.id, b'GT')

    def test_dos_type(self):
        self.image.bam_block.data[2] = 0x41
        self.assertEqual(self.image.dos_type, ord('A'))


class TestD64ImageWrite(unittest.TestCase):

    def setUp(self):
        self.image = D64Image(None)
        self.image.bam_block = MockBlock()

    def test_write_dos_version(self):
        self.image.dos_version = 0x64
        self.assertEqual(self.image.dos_version, 0x64)

    def test_write_name(self):
        self.image.name = b'EXAMPLE'
        self.assertEqual(self.image.name, b'EXAMPLE')
        self.image.name = b'SHORT'
        self.assertEqual(self.image.bam_block.data[0x90:0x96], b'SHORT\xa0')
        self.image.name = b'VERY LONG EXAMPLE'
        self.assertEqual(self.image.name, b'VERY LONG EXAMPL')
        self.assertNotEqual(self.image.bam_block.data[0xa0], 0x45)

    def test_write_id(self):
        self.image.id = b'EX'
        self.assertEqual(self.image.id, b'EX')
        with self.assertRaises(ValueError):
            self.image.id = b'LONG'

    def test_write_dos_type(self):
        self.image.dos_type = 0x77
        self.assertEqual(self.image.dos_type, 0x77)
        self.assertEqual(self.image.bam_block.data[2], 0x77)
        self.assertEqual(self.image.bam_block.data[0xa6], 0x77)


class TestD64ImageBlocks(unittest.TestCase):

    def setUp(self):
        self.image = D64Image(None)
        self.image.bam_block = MockBlock()

    def test_max_sectors(self):
        self.assertEqual(self.image.max_sectors(1), 21)
        self.assertEqual(self.image.max_sectors(5), 21)
        self.assertEqual(self.image.max_sectors(10), 21)
        self.assertEqual(self.image.max_sectors(15), 21)
        self.assertEqual(self.image.max_sectors(20), 19)
        self.assertEqual(self.image.max_sectors(25), 18)
        self.assertEqual(self.image.max_sectors(30), 18)
        self.assertEqual(self.image.max_sectors(35), 17)

    def test_max_sectors_bad(self):
        with self.assertRaises(ValueError):
            self.image.max_sectors(0)
        with self.assertRaises(ValueError):
            self.image.max_sectors(36)

    def test_block_start(self):
        self.assertEqual(self.image.block_start(1, 0), 0)
        self.assertEqual(self.image.block_start(1, 1), 256)
        self.assertEqual(self.image.block_start(1, 20), 5120)
        self.assertEqual(self.image.block_start(2, 0), 5376)
        self.assertEqual(self.image.block_start(35, 16), 174592)

    def test_block_start_bad(self):
        with self.assertRaises(ValueError):
            self.image.block_start(0, 0)
        with self.assertRaises(ValueError):
            self.image.block_start(1, 21)


if __name__ == '__main__':
    unittest.main()



import unittest

from unittest.mock import Mock

from d64.partition import Partition

from test.mock_block import MockBlock


class TestCreatePartition(unittest.TestCase):

    def setUp(self):
        self.entry = Mock()
        self.image = Mock()
        self.image.get_free_entry.return_value = self.entry
        self.part = Partition(self.image, name=b'test')

    def test_create(self):
        start_block = MockBlock(self.image, 1, 0)
        self.entry.partition_blocks.return_value = iter([start_block])
        self.part.create(start_block, 1)
        self.assertEqual(self.entry.name, b'test')
        self.image.bam.set_allocated.assert_called_once()

    def test_create_bad_block(self):
        start_block = MockBlock(self.image, 1, 0)

        def part_blocks():
            yield start_block
            raise ValueError("mocked bad sector")

        self.entry.partition_blocks = part_blocks
        with self.assertRaises(ValueError):
            self.part.create(start_block, 2)
        self.image.free_block.assert_called_once()
        self.entry.reset.assert_called_once()
        self.assertIsNone(self.part.entry)

    def test_create_exists(self):
        mock_entry = Mock()
        mock_entry.name.return_value = b'test'
        part = Partition(None, entry=mock_entry)
        with self.assertRaises(FileExistsError):
            part.create(None, None)



from .block import Block


class SideSector(Block):
    """Relative file side sector block."""
    MAX_SIDE_SECTORS = 6
    MAX_DATA_LINKS = 120

    @property
    def number(self):
        """Return this side sector block number (0-5)."""
        return self.get(2)

    @property
    def record_len(self):
        """Return record length."""
        return self.get(3)

    @number.setter
    def number(self, number):
        """Modify side sector block number."""
        if number < 0 or number >= self.MAX_SIDE_SECTORS:
            raise ValueError("Invalid side sector index, %d" % number)
        self.set(2, number)

    @record_len.setter
    def record_len(self, rec_len):
        """Modify side sector record length."""
        self.set(3, rec_len)

    def all_side_sectors(self):
        """Return an array of all side sectors track & sector."""
        ss_bin = self.get(4, 0x10)
        return [(ss_bin[i], ss_bin[i+1]) for i in range(0, self.MAX_SIDE_SECTORS*2, 2) if ss_bin[i]]

    def all_data_blocks(self):
        """Return an array of all data blocks in this side sector."""
        links_bin = self.get(0x10, 0x100)
        return [(links_bin[i], links_bin[i+1]) for i in range(0, 0xf0, 2) if links_bin[i]]

    def clear_side_sectors(self):
        """Zero out all side sector links."""
        self.set(4, bytes(self.MAX_SIDE_SECTORS*2))

    def clear_data_blocks(self):
        """Zero out all data block links."""
        self.set(0x10, bytes(0xf0))
        self.data_size = self.MAX_SIDE_SECTORS*2+2

    def add_side_sector(self, other):
        """Append a new side sector to this side sector's list."""
        self.set(4+other.number*2, bytes((other.track, other.sector)))

    def add_data_block(self, block):
        """Append a new data block to this side sector."""
        self.set(self.data_size+2, bytes((block.track, block.sector)))
        self.data_size += 2

    def set_peers(self, ss_ts_list):
        """Set the list of all side sectors."""
        padded_ss_list = ss_ts_list+[(0, 0)]*(self.MAX_SIDE_SECTORS-len(ss_ts_list))
        self.set(4, bytes([item for sublist in padded_ss_list for item in sublist]))

    def set_data_blocks(self, data_list):
        """Set the list of data block links."""
        padded_data_list = [(d.track, d.sector) for d in data_list]+[(0, 0)]*(self.MAX_DATA_LINKS-len(data_list))
        self.set(0x10, bytes([item for sublist in padded_data_list for item in sublist]))
        if self.is_final:
            self.data_size = len(data_list)*2+self.MAX_SIDE_SECTORS*2+2



import unittest

from unittest.mock import patch, Mock

from d64.dos_image import DOSImage

from test.mock_bam import MockBAM
from test.mock_block import MockBlock


class TestDOSImageAlloc(unittest.TestCase):

    def setUp(self):
        self.image = DOSImage(None)
        self.image.DIR_TRACK = 18
        self.image.DIR_SECTOR = 1
        self.image.MAX_TRACK = 35
        self.image.TRACK_SECTOR_MAX = ((21, (1, 35)), )
        self.image.bam = MockBAM(self.image.TRACK_SECTOR_MAX)

    def test_alloc_first_below(self):
        self.image.bam.fill_entry(17)
        block = self.image.alloc_first_block()
        self.assertIsNotNone(block)
        self.assertEqual(block.track, 17)
        self.assertEqual(block.sector, 0)
        block = self.image.alloc_first_block()
        self.assertIsNotNone(block)
        self.assertEqual(block.track, 17)
        self.assertEqual(block.sector, 1)

    def test_alloc_first_above(self):
        self.image.bam.clear_entry(17)
        self.image.bam.fill_entry(19)
        block = self.image.alloc_first_block()
        self.assertIsNotNone(block)
        self.assertEqual(block.track, 19)
        self.assertEqual(block.sector, 0)
        block = self.image.alloc_first_block()
        self.assertIsNotNone(block)
        self.assertEqual(block.track, 19)
        self.assertEqual(block.sector, 1)

    def test_alloc_first_full(self):
        for t in range(1, 36):
            self.image.bam.clear_entry(t)
        self.image.bam.fill_entry(18)
        self.assertIsNone(self.image.alloc_first_block())

    def test_alloc_next_interleave(self):
        self.image.bam.fill_entry(17)
        block = self.image._alloc_next_block(17, 0, 10)
        self.assertIsNotNone(block)
        self.assertEqual(block.track, 17)
        self.assertEqual(block.sector, 10)
        block = self.image._alloc_next_block(17, 0, 10)
        self.assertIsNotNone(block)
        self.assertEqual(block.track, 17)
        self.assertEqual(block.sector, 11)
        block = self.image._alloc_next_block(17, 16, 10)
        self.assertIsNotNone(block)
        self.assertEqual(block.track, 17)
        self.assertEqual(block.sector, 4)
        block = self.image._alloc_next_block(17, 11, 10)
        self.assertIsNotNone(block)
        self.assertEqual(block.track, 17)
        self.assertEqual(block.sector, 0)

    def test_alloc_diff_track(self):
        self.image.bam.clear_entry(17)
        self.image.bam.fill_entry(16)
        block = self.image._alloc_next_block(17, 0, 10)
        self.assertIsNotNone(block)
        self.assertEqual(block.track, 16)
        self.assertEqual(block.sector, 10)

    def test_alloc_above(self):
        for t in range(1, 18):
            self.image.bam.clear_entry(t)
        self.image.bam.fill_entry(19)
        block = self.image._alloc_next_block(17, 0, 10)
        self.assertIsNotNone(block)
        self.assertEqual(block.track, 19)
        self.assertEqual(block.sector, 10)

    def test_alloc_dir(self):
        for t in range(1, 36):
            self.image.bam.clear_entry(t)
        self.image.bam.fill_entry(18)
        block = self.image._alloc_next_block(17, 0, 10)
        self.assertIsNotNone(block)
        self.assertEqual(block.track, 18)
        self.assertEqual(block.sector, 10)

    def test_alloc_full(self):
        for t in range(1, 36):
            self.image.bam.clear_entry(t)
        block = self.image._alloc_next_block(17, 0, 10)
        self.assertIsNone(block)


class MockImage(DOSImage):
    DIR_TRACK = 18
    DIR_SECTOR = 1
    DIR_INTERLEAVE = 3

    def __init__(self, filename):
        super().__init__(filename)
        self._alloc_first_block_return = []
        self._alloc_next_block_return = []
        self.bam = Mock()
        self._free_called_for = []

    def alloc_first_block(self):
        if self._alloc_first_block_return:
            return self._alloc_first_block_return.pop(0)
        return None

    def alloc_next_block(self, _, __, directory=False):
        if self._alloc_next_block_return:
            return self._alloc_next_block_return.pop(0)
        return None

    def free_block(self, block):
        self._free_called_for.append(block)


class TestDirEntry(unittest.TestCase):
    def setUp(self):
        self.free_dir_entry = bytes(64)
        self.in_use_dir_entry = b'\x00\xff\x82\x0A\x14\x46\x49\x47\x48\x54\x45\x52\x20\x52\x41\x49' \
                                b'\x44\xA0\xA0\xA0\xA0\x00\x00\x00\x00\x00\x00\x00\x00\x00\xE7\x01'

    def test_deleted_entry_exists(self):
        image = MockImage(None)
        MockBlock.BLOCK_FILL = self.free_dir_entry * 8
        with patch('d64.dos_image.Block', new=MockBlock):
            entry = image.get_free_entry()
        self.assertIsNotNone(entry)
        self.assertEqual(entry.block.track, 18)
        self.assertEqual(entry.block.sector, 1)

    def test_from_new_block(self):
        image = MockImage(None)
        MockBlock.BLOCK_FILL = self.in_use_dir_entry * 8
        image._alloc_next_block_return = [MockBlock(image, 18, 4)]
        with patch('d64.dos_image.Block', new=MockBlock):
            entry = image.get_free_entry()
        self.assertIsNotNone(entry)
        self.assertEqual(entry.block.track, 18)
        self.assertEqual(entry.block.sector, 4)

    def test_no_free_next(self):
        image = MockImage(None)
        MockBlock.BLOCK_FILL = self.in_use_dir_entry * 8
        image._alloc_next_block_return = []
        with patch('d64.dos_image.Block', new=MockBlock):
            entry = image.get_free_entry()
        self.assertIsNone(entry)

    def tearDown(self):
        MockBlock.BLOCK_FILL = bytes(64) * 4


class TestChains(unittest.TestCase):

    def test_clone_chain(self):
        image = MockImage(None)
        image._alloc_next_block_return = [MockBlock(image, 20, 2), MockBlock(image, 20, 7), MockBlock(image, 20, 11)]
        MockBlock.BLOCK_FILL = b'\x99\x88\x77\x66\x55\x44\x33\x22' * 64
        block1 = MockBlock()
        block2 = MockBlock()
        block3 = MockBlock()
        block3.data_size = 176
        block2.set_next_block(block3)
        block1.set_next_block(block2)
        new_block1 = image.clone_chain(block1)
        self.assertIsNotNone(new_block1)
        self.assertEqual(new_block1.sector, 2)
        self.assertIsNotNone(new_block1.next_block())
        self.assertEqual(new_block1.next_block().sector, 7)
        self.assertIsNotNone(new_block1.next_block().next_block())
        self.assertEqual(new_block1.next_block().next_block().sector, 11)
        self.assertTrue(new_block1.next_block().next_block().is_final)

    def tearDown(self):
        MockBlock.BLOCK_FILL = bytes(64) * 4


class TestValidImage(unittest.TestCase):

    def test_valid_image(self):
        class TestImage(DOSImage):
            IMAGE_SIZES = (1234, )

        mock_stat = Mock()
        mock_stat.st_size = 1234
        mock_path = Mock()
        mock_path.stat.return_value = mock_stat
        self.assertTrue(TestImage.valid_image(mock_path))
        mock_stat.st_size = 4321
        self.assertFalse(TestImage.valid_image(mock_path))


if __name__ == '__main__':
    unittest.main()



class ConsistencyError(Exception):
    """Raised when an inconsistency in the image is detected."""
    pass


class DiskFullError(Exception):
    """Raised by a write operation when no blocks are free."""
    pass


class FileIndexError(Exception):
    """Raised when no more side sectors can added."""
    pass



import unittest

from unittest.mock import patch

from d64.side_sector import SideSector
from d64.super_side_sector import SuperSideSector

from test.mock_block import MockBlock


class TestSuperSideSector(unittest.TestCase):

    def setUp(self):
        p = patch.object(SuperSideSector, '__bases__', (MockBlock,))
        with p:
            p.is_local = True
            self.super_side = SuperSideSector(None, None, None)

    def test_clear_side_sectors(self):
        p1 = patch.object(SuperSideSector, '__bases__', (MockBlock,))
        with p1:
            p1.is_local = True
            p2 = patch.object(SideSector, '__bases__', (MockBlock,))
            with p2:
                p2.is_local = True
                self.super_side.data[3:0xff] = b'\xee\x22' * 126
                self.super_side.clear_side_sectors()
                self.assertEqual(len(self.super_side.all_side_sectors()), 0)
                self.assertEqual(self.super_side.super_id, 0xfe)

    def test_add_side_sector(self):
        mock_side = MockBlock(None, 30, 7)
        p1 = patch.object(SuperSideSector, '__bases__', (MockBlock,))
        with p1:
            p1.is_local = True
            p2 = patch.object(SideSector, '__bases__', (MockBlock,))
            with p2:
                p2.is_local = True
                self.super_side.add_side_sector(mock_side)
                self.assertEqual(len(self.super_side.all_side_sectors()), 1)

    def test_side_sector(self):
        mock_side = MockBlock(None, 30, 7)
        p1 = patch.object(SuperSideSector, '__bases__', (MockBlock,))
        with p1:
            p1.is_local = True
            p2 = patch.object(SideSector, '__bases__', (MockBlock,))
            with p2:
                p2.is_local = True
                self.super_side.set_side_sector(4, mock_side)
                self.assertEqual(self.super_side.side_sector(4).track, 30)
                with self.assertRaises(ValueError):
                    self.super_side.side_sector(130)
                with self.assertRaises(ValueError):
                    self.super_side.set_side_sector(130, mock_side)



import unittest

from unittest.mock import Mock

from d64.exceptions import DiskFullError
from d64.file import File

from test.mock_block import MockBlock


class TestFileRead(unittest.TestCase):

    def setUp(self):
        MockBlock.BLOCK_FILL = b'\x00\x01\x02\x03' * 64
        mock_entry = Mock()
        mock_entry.first_block.return_value = MockBlock()
        self.file = File(mock_entry, 'r')
        self.file.block.data_size = 254

    def test_read_small(self):
        data = self.file.read(10)
        self.assertEqual(len(data), 10)
        self.assertEqual(data, b'\x02\x03\x00\x01\x02\x03\x00\x01\x02\x03')
        data = self.file.read(4)
        self.assertEqual(len(data), 4)
        self.assertEqual(data, b'\x00\x01\x02\x03')

    def test_read_past_end(self):
        self.file.block.data_size = 10
        data = self.file.read(20)
        self.assertEqual(len(data), 10)
        self.assertEqual(data, b'\x02\x03\x00\x01\x02\x03\x00\x01\x02\x03')
        data = self.file.read(20)
        self.assertEqual(len(data), 0)

    def test_read_multi_block(self):
        block2 = MockBlock()
        block2._set_data(b'\x10\x11\x12\x13' * 64)
        self.file.block.set_next_block(block2)
        data = self.file.read(303)
        self.assertEqual(len(data), 303)
        expected = b'\x02\x03'+b'\x00\x01\x02\x03'*63+b'\x12\x13'+b'\x10\x11\x12\x13'*11+b'\x10\x11\x12'
        self.assertEqual(data, expected)

    def test_read_boundary(self):
        block2 = MockBlock()
        block2._set_data(b'\x10\x11\x12\x13' * 64)
        self.file.block._set_data(b'\x00\x01\x02\x03' * 64)
        self.file.block.set_next_block(block2)
        data = self.file.read(254)
        self.assertEqual(len(data), 254)
        expected = b'\x02\x03'+b'\x00\x01\x02\x03'*63
        self.assertEqual(data, expected)
        data = self.file.read(49)
        self.assertEqual(len(data), 49)
        expected = b'\x12\x13'+b'\x10\x11\x12\x13'*11+b'\x10\x11\x12'
        self.assertEqual(data, expected)

    def tearDown(self):
        MockBlock.BLOCK_FILL = bytes(64) * 4


class TestFileWrite(unittest.TestCase):

    def setUp(self):
        mock_image = Mock()
        mock_image.alloc_first_block.return_value = MockBlock(mock_image)
        mock_image.alloc_next_block.return_value = MockBlock(mock_image)
        self.entry = Mock()
        self.entry.size = 0
        self.entry.block = MockBlock(mock_image)
        self.file = File(self.entry, 'w')

    def test_write_small(self):
        self.assertEqual(self.file.write(b'abcdefg\r'), 8)
        self.assertEqual(self.file.block.data[2:10], b'abcdefg\r')
        self.assertEqual(self.file.write(b'tuvwxyz\r'), 8)
        self.assertEqual(self.file.block.data[2:18], b'abcdefg\rtuvwxyz\r')

    def test_write_multi_block(self):
        self.assertEqual(self.file.write(b'\x00\x01\x02\x03' * 70), 280)
        block1 = self.entry.set_first_block.call_args_list[0][0][0]
        self.assertEqual(block1.data[2:], b'\x00\x01\x02\x03'*63+b'\x00\x01')
        block2 = block1.next_block()
        self.assertIsNotNone(block2)
        self.assertEqual(block2.data[2:28], b'\x02\x03'+b'\x00\x01\x02\x03'*6)
        self.assertEqual(self.file.entry.size, 2)

    def test_write_multi_full(self):
        self.file.image.alloc_next_block.return_value = None
        with self.assertRaises(DiskFullError):
            self.file.write(b'\x00\x01\x02\x03' * 70)

    def test_write_empty(self):
        self.file.__exit__(None, None, None)
        self.assertEqual(self.file.block.data_size, 1)
        self.assertEqual(self.file.block.data[2], 0x0d)


if __name__ == '__main__':
    unittest.main()



import argparse
import sys
import tempfile

import cmd2

from d64 import Block, DiskImage


class BlockRef:
    def __init__(self, ts):
        if ':' in ts:
            track_s, sector_s = ts.split(':', 1)
            self.track = int(track_s)
            self.sector = int(sector_s)
        else:
            raise ValueError("Invalid track/sector, "+ts)


def as_text(block, start=0, end=256):
    chunk = 16
    while start < end:
        span = chunk - (start % chunk)
        fmt = '{:02x} '*span
        yield fmt.format(*[b for b in block.get(start, start+span)])
        start += span


def update_from_file(block, fileh, start=0):
    for line in fileh:
        line = line.rstrip()
        new_data = bytes([int(b, 16) for b in line.split()])
        block.set(start, new_data)
        start += len(new_data)


class D64Edit(cmd2.Cmd):
    def __init__(self, image):
        super().__init__(allow_cli_args=False)
        self.prompt = '(edit) '
        self.image = image
        self.block = Block(image, image.DIR_TRACK, 0)

    def show_block(self):
        if self.block.is_final:
            try:
                data_size = self.block.data_size
            except ValueError:
                data_size = '-'
            self.poutput("Block: {}, data size: {}\n".format(self.block, data_size))
        else:
            try:
                next_block = str(self.block.next_block())
            except ValueError:
                next_block = '-'
            self.poutput("Block: {}, next block: {}, offset: {:08x}\n".format(self.block, next_block, self.block.start))
        for line in as_text(self.block):
            self.poutput(line)

    def do_directory(self, args):
        """Display all directory entries"""
        for path in self.image.iterdir(include_deleted=True):
            line = "name: {!s}, type: {:02x}, first block: {}".format(path.entry.name, path.entry._file_type(), path.entry.first_block())
            self.poutput(line)

    def do_bam(self, args):
        """Display Block Availability Map"""
        for track in range(1, self.image.MAX_TRACK+1):
            total, bits = self.image.bam.get_entry(track)
            self.poutput("Track {:2d}  {} {:2d}".format(track, bits, total))

    block_parser = argparse.ArgumentParser()
    block_parser.add_argument('block', type=BlockRef, nargs=(0, 1), metavar='TRACK:SECTOR', help="track and sector")

    @cmd2.with_argparser(block_parser)
    def do_block(self, args):
        """Display contents of a block"""
        if args.block:
            self.block = Block(self.image, args.block.track, args.block.sector)

        self.show_block()

    next_parser = argparse.ArgumentParser()
    next_parser.add_argument('block', type=BlockRef, nargs=(0, 1), metavar='TRACK:SECTOR', help="track and sector")

    @cmd2.with_argparser(next_parser)
    def do_next(self, args):
        """Follow link to next block"""
        if args.block:
            self.block = Block(self.image, args.block.track, args.block.sector)

        next_block = self.block.next_block()
        if next_block:
            self.block = next_block
            self.show_block()
        else:
            self.poutput("Final block")

    edit_block_parser = argparse.ArgumentParser()
    edit_block_parser.add_argument('block', type=BlockRef, nargs=(0, 1), metavar='TRACK:SECTOR', help="track and sector")

    @cmd2.with_argparser(edit_block_parser)
    def do_edit_block(self, args):
        """Edit contents of a block"""
        if args.block:
            self.block = Block(self.image, args.block.track, args.block.sector)

        with tempfile.NamedTemporaryFile(mode='w+', buffering=1, prefix="edit_block-{!s}-".format(self.block)) as fileh:
            for line in as_text(self.block):
                print(line, file=fileh)
            fileh.seek(0)
            if hasattr(self, 'run_editor'):
                self.run_editor(fileh.name)
            else:
                self._run_editor(fileh.name)
            update_from_file(self.block, fileh)

    edit_entry_parser = argparse.ArgumentParser()
    edit_entry_parser.add_argument('name', help="Directory entry name")

    @cmd2.with_argparser(edit_entry_parser)
    def do_edit_entry(self, args):
        """Edit contents of a directory entry"""
        entry = self.image.path(args.name.encode('ascii')).entry

        with tempfile.NamedTemporaryFile(mode='w+', buffering=1, prefix="edit_entry-") as fileh:
            for line in as_text(entry.block, start=entry.entry_offset, end=entry.entry_offset+entry.ENTRY_SIZE):
                print(line, file=fileh)
            fileh.seek(0)
            if hasattr(self, 'run_editor'):
                self.run_editor(fileh.name)
            else:
                self._run_editor(fileh.name)
            update_from_file(entry.block, fileh, start=entry.entry_offset)

    set_next_parser = argparse.ArgumentParser()
    set_next_parser.add_argument('block', type=BlockRef, metavar='TRACK:SECTOR', help="track and sector to link to")

    @cmd2.with_argparser(set_next_parser)
    def do_set_next(self, args):
        """Update link to next block"""
        next_block = Block(self.image, args.block.track, args.block.sector)
        self.block.set_next_block(next_block)

    set_size_parser = argparse.ArgumentParser()
    set_size_parser.add_argument('size', type=int, help="data size")

    @cmd2.with_argparser(set_size_parser)
    def do_set_size(self, args):
        """Update data size of final block"""
        self.block.data_size = args.size

    alloc_block_parser = argparse.ArgumentParser()
    alloc_block_parser.add_argument('block', type=BlockRef, metavar='TRACK:SECTOR', help="track and sector")

    @cmd2.with_argparser(alloc_block_parser)
    def do_alloc_block(self, args):
        """Allocate a block in the BAM"""
        self.image.bam.set_allocated(args.block.track, args.block.sector)

    free_block_parser = argparse.ArgumentParser()
    free_block_parser.add_argument('block', type=BlockRef, metavar='TRACK:SECTOR', help="track and sector")

    @cmd2.with_argparser(free_block_parser)
    def do_free_block(self, args):
        """Free a block in the BAM"""
        self.image.bam.set_free(args.block.track, args.block.sector)

    def do_alloc_first(self, args):
        """Allocate the first block in a file"""
        new_block = self.image.alloc_first_block()
        if new_block:
            self.poutput("New block: {}".format(new_block))
            self.block = new_block
        else:
            self.pwarning("Disk full")

    alloc_next_parser = argparse.ArgumentParser()
    alloc_next_parser.add_argument('block', type=BlockRef, nargs=(0, 1), metavar='TRACK:SECTOR', help="track and sector")
    alloc_next_parser.add_argument('--directory', action='store_true', help="allocate from directory track")

    @cmd2.with_argparser(alloc_next_parser)
    def do_alloc_next(self, args):
        """Allocate next block for a file"""
        if args.block:
            self.block = Block(self.image, args.block.track, args.block.sector)

        new_block = self.image.alloc_next_block(self.block.track, self.block.sector, directory=args.directory)
        if new_block:
            self.poutput("New block: {}".format(new_block))
            self.block = new_block
        else:
            self.pwarning("Disk full")


def main():
    parser = argparse.ArgumentParser(description="Edit Commodore disk images")
    parser.add_argument('image', help="image file name")
    args = parser.parse_args()

    with DiskImage(args.image, mode='w') as image:
        e = D64Edit(image)
        ret = e.cmdloop()

    sys.exit(ret)



import logging

from .block import Block
from .exceptions import DiskFullError, FileIndexError
from .file import File
from .side_sector import SideSector

log = logging.getLogger(__name__)


class RelativeFile(File):
    """Read and write access to relative files."""
    def get_first_side_sector(self, data_block):
        """Allocate side sector and initialize it."""
        block = self.image.alloc_next_block(data_block.track, data_block.sector)
        if block is None:
            return None

        side_sector = SideSector(self.image, block.track, block.sector)
        log.debug("Allocated first side sector %s", side_sector)
        side_sector.number = 0
        side_sector.record_len = self.entry.record_len
        side_sector.clear_side_sectors()
        side_sector.add_side_sector(side_sector)
        side_sector.clear_data_blocks()
        return side_sector

    def get_next_side_sector(self, data_block):
        """Allocate the next side sector."""
        if self.side_sector.number+1 == SideSector.MAX_SIDE_SECTORS:
            raise FileIndexError()

        block = self.image.alloc_next_block(data_block.track, data_block.sector)
        if block is None:
            raise DiskFullError()

        side_sector = SideSector(self.image, block.track, block.sector)
        side_sector.number = self.side_sector.number+1
        log.debug("Allocated side sector %d %s", side_sector.number, side_sector)
        side_sector.record_len = self.entry.record_len
        side_sector.set_peers(self.side_sector.all_side_sectors())
        side_sector.clear_data_blocks()
        self.side_sector.add_side_sector(side_sector)
        self.side_sector.set_next_block(side_sector)

        # update the peer list of all side sectors (including the new one)
        all_peers = self.side_sector.all_side_sectors()
        for t, s in all_peers:
            ss = SideSector(self.image, t, s)
            ss.set_peers(all_peers)

        self.side_sector = side_sector
        self.entry.size += 1

    def get_first_block(self):
        """Return the first empty block for a file write."""
        data_block = super().get_first_block()
        side_sector = self.get_first_side_sector(data_block)
        if side_sector is None:
            self.image.free_block(data_block)
            raise DiskFullError()

        side_sector.add_data_block(data_block)
        self.entry.side_sector_ts = (side_sector.track, side_sector.sector)
        self.side_sector = side_sector
        self.entry.size += 1
        return data_block

    def get_new_block(self):
        """Get a new empty block for a file write."""
        data_block = super().get_new_block()
        if len(self.side_sector.all_data_blocks()) == SideSector.MAX_DATA_LINKS:
            # current side sector is full, allocate a new one
            self.get_next_side_sector(data_block)

        self.side_sector.add_data_block(data_block)
        return data_block

    def read_record(self):
        """Read a complete record from a file."""
        return self.read(self.entry.record_len)

    def write(self, data):
        """Write one or more records to a file."""
        ret = 0
        while data:
            chunk = data[:self.entry.record_len]
            ret += super().write(chunk.ljust(self.entry.record_len, b'\x00'))
            data = data[self.entry.record_len:]
        return ret

    def close(self):
        """Close file."""
        log.debug("Closing %s", self.entry.name)
        # pad rest of data block with empty records
        while self.block is None or Block.SECTOR_SIZE-(self.block.data_size+2) >= self.entry.record_len:
            self.write(b'\xff')



import logging
import sys

import petscii_codecs

from d64 import DiskImage


logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)

with DiskImage(sys.argv[1]) as image:
    for line in image.directory(encoding='petscii-c64en-uc'):
        print(line)

    print()
    subdir = image.subdirectory(sys.argv[2].encode('petscii-c64en-uc'))
    for line in subdir.directory(encoding='petscii-c64en-uc'):
        print(line)



from .block import Block


class D80D82BAMBlock(Block):
    """BAM block (8050/8250)."""

    @property
    def dos_type(self):
        return self.get(2)

    @dos_type.setter
    def dos_type(self, dtype):
        self.set(2, dtype)

    @property
    def track_range(self):
        tr = tuple(self.get(4, 6))
        return tr[0], tr[1]-1

    @track_range.setter
    def track_range(self, tr):
        start, end = tr
        self.set(4, bytes((start, end+1)))



import logging

from .block import Block
from .exceptions import DiskFullError

log = logging.getLogger(__name__)


class File(object):
    """Read and write access to files."""
    def __init__(self, entry, mode):
        log.debug("Opening %s for %s", entry.name, mode)
        self.entry = entry
        if mode[0] == 'r':
            self.block = entry.first_block()
            self.read_pos = 2
        else:
            self.image = entry.block.image
            self.block = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, _, __):
        if exc_type is None:
            self.close()

    def get_first_block(self):
        """Return the first empty block for a file write."""
        next_block = self.image.alloc_first_block()
        if next_block is None:
            raise DiskFullError()
        next_block.data_size = 0
        self.entry.set_first_block(next_block)
        self.entry.size = 1
        log.debug("Allocated first data block %s", next_block)
        return next_block

    def get_new_block(self):
        """Get a new empty block for a file write."""
        next_block = self.image.alloc_next_block(self.block.track, self.block.sector)
        if next_block is None:
            raise DiskFullError()
        log.debug("Allocated data block %s", next_block)
        next_block.data_size = 0
        self.block.set_next_block(next_block)
        self.entry.size += 1
        return next_block

    def read(self, count=-1):
        """Read bytes from file."""
        ret = b''

        while count:
            remaining = self.block.data_size-(self.read_pos-2)
            if remaining == 0:
                break

            # read as much as is wanted from the current block
            length = remaining if count == -1 else min(count, remaining)
            ret += self.block.get(self.read_pos, self.read_pos+length)
            self.read_pos += length
            if count != -1:
                count -= length

            if self.block.is_final:
                # no more blocks, end of file
                break

            if self.read_pos == Block.SECTOR_SIZE:
                # end of block, move on to the next block
                self.block = self.block.next_block()
                self.read_pos = 2

        return ret

    def write(self, data):
        """Write data to a file."""
        written = 0

        if self.block is None:
            # allocate first block
            self.block = self.get_first_block()

        while data:
            remaining_space = Block.SECTOR_SIZE-(self.block.data_size+2)
            if remaining_space:
                length = min(remaining_space, len(data))
                self.block.set(self.block.data_size+2, data[:length])
                self.block.data_size += length
                written += length
                data = data[length:]
            else:
                # end of block, append a new one
                self.block = self.get_new_block()

        return written

    def close(self):
        """Close file."""
        log.debug("Closing %s", self.entry.name)
        if self.block is None:
            # file open for write, no data written
            # DOS files cannot be empty, they must contain at least one byte
            self.write(b'\r')



import math

from .block import Block
from .d81_bam import D81BAM
from .d81_bam_block import D81BAMBlock
from .d81_image import D81Image


class Subdirectory(D81Image):

    def __init__(self, parent, entry):
        self.parent = parent
        self.entry = entry
        self.map = parent.map
        self.writeable = parent.writeable
        self.MIN_TRACK = self.DIR_TRACK = entry.first_block().track
        self.MAX_TRACK = self.MIN_TRACK+math.ceil(entry.size/self.SECTORS_PER_TRACK)-1
        self.bam = D81BAM(self)
        self.header_block = Block(self, self.DIR_TRACK, 0)
        self.side_a_bam_block = D81BAMBlock(self, self.DIR_TRACK, 1)
        self.side_b_bam_block = D81BAMBlock(self, self.DIR_TRACK, 2)
        self.dir_block = Block(self, self.DIR_TRACK, self.DIR_SECTOR)

    def open(self, mode):
        return

    def close(self):
        return

    def is_formatted(self):
        """Return `True` if a subdirectory is valid."""
        return self.dos_type == self.DOS_FORMAT and self.dos_version == self.DOS_VERSION

    @classmethod
    def create(cls, parent, entry, disk_name, disk_id):
        """Format a partition to create an empty subdirectory."""
        if entry.start_ts[1] != 0:
            raise ValueError("Partition does not start at sector 0")
        if entry.size < 120:
            raise ValueError("Partition is less than 120 blocks")
        if entry.size % cls.SECTORS_PER_TRACK:
            raise ValueError("Partition is not a multiple of 40 blocks")

        for block in entry.partition_blocks():
            block.set(0, bytes(block.SECTOR_SIZE))
        subdir = cls(parent, entry)
        subdir._create(disk_name, disk_id)

    def __str__(self):
        return "Subdirectory({})".format(self.entry.name)



import unittest
from unittest.mock import Mock

from d64.d81_bam import D81BAM

from test.mock_block import MockBlock


class TestBAM(unittest.TestCase):

    def setUp(self):
        self.image = Mock()
        self.image.MIN_TRACK = 1
        self.image.MAX_TRACK = 80
        self.image.side_a_bam_block = MockBlock()
        self.image.side_b_bam_block = MockBlock()
        self.image.bam = D81BAM(self.image)

    def test_get_entry(self):
        self.image.side_a_bam_block.data[0x1c:0x22] = b'\x17\xc9\x6e\x5b\x6f\x85'
        entry = self.image.bam.get_entry(3)
        self.assertEqual(entry, (23, '1001001101110110110110101111011010100001'))
        self.image.side_b_bam_block.data[0x34:0x3a] = b'\x17\xc9\x6e\x5b\x6f\x85'
        entry = self.image.bam.get_entry(47)
        self.assertEqual(entry, (23, '1001001101110110110110101111011010100001'))

    def test_set_entry(self):
        self.image.bam.set_entry(3, 26, '1101101101110001101010111011101010110111')
        self.assertEqual(self.image.side_a_bam_block.data[0x1c:0x22], b'\x1a\xdb\x8e\xd5]\xed')
        self.image.bam.set_entry(47, 26, '1101101101110001101010111011101010110111')
        self.assertEqual(self.image.side_b_bam_block.data[0x34:0x3a], b'\x1a\xdb\x8e\xd5]\xed')


if __name__ == '__main__':
    unittest.main()



import unittest

from unittest.mock import Mock

from d64.block import Block


class TestBlock(unittest.TestCase):

    def setUp(self):
        self.image = Mock()
        self.image.block_start.return_value = 0
        self.image.map = bytes(Block.SECTOR_SIZE)

    def test_track_sector(self):
        block = Block(self.image, 10, 20)
        self.assertEqual(block.track, 10)
        self.assertEqual(block.sector, 20)

    def test_is_final(self):
        block = Block(self.image, None, None)
        self.image.map = b'\x00'
        self.assertTrue(block.is_final)
        self.image.map = b'\x12'
        self.assertFalse(block.is_final)

    def test_data_size(self):
        block = Block(self.image, None, None)
        self.image.map = b'\x00\xff'
        self.assertEqual(block.data_size, 254)
        self.image.map = bytearray(2)
        block.data_size = 42
        self.assertEqual(self.image.map[0], 0)
        self.assertEqual(self.image.map[1], 43)
        with self.assertRaises(ValueError):
            block.data_size = 400

    def test_next_block(self):
        block = Block(self.image, None, None)
        self.image.map = b'\x1a\x17'
        next_block = block.next_block()
        self.assertEqual(next_block.track, 26)
        self.assertEqual(next_block.sector, 23)
        self.image.map = bytearray(2)
        block2 = Block(self.image, 9, 14)
        block.set_next_block(block2)
        self.assertEqual(self.image.map[0], 9)
        self.assertEqual(self.image.map[1], 14)

    def test_data_access(self):
        block = Block(self.image, None, None)
        self.image.map = b'\x00\x10\x11\x12\x13\x14\x15'
        self.assertEqual(block.get(2, 6), b'\x11\x12\x13\x14')

    def test_equal(self):
        blocks = [Block(self.image, 3, 4), Block(self.image, 5, 6)]
        self.assertIn(Block(self.image, 5, 6), blocks)
        self.assertNotIn(Block(self.image, 7, 8), blocks)


if __name__ == '__main__':
    unittest.main()



import struct

from .block import Block
from .relative_file import RelativeFile


class DirEntry(object):
    ENTRY_SIZE = 0x20
    FTYPE_STR = ('DEL', 'SEQ', 'PRG', 'USR', 'REL', '???', '???', '???')

    def __init__(self, block, entry_offset):
        self.block = block
        self.entry_offset = entry_offset

    def _file_type(self):
        return self.block.get(self.entry_offset+2)

    def first_block(self):
        """Return the first block containing file data."""
        if self.start_ts[0] == 0:
            return None
        return Block(self.block.image, *self.start_ts)

    def set_first_block(self, block):
        """Set the first block containing file data."""
        self.start_ts = (block.track, block.sector)

    def reset(self):
        """Reset an entry for reuse."""
        self.size = 0
        self.start_ts = (0, 0)
        self.side_sector_ts = (0, 0)

    def free_blocks(self):
        """Free any data blocks and side sectors."""
        try:
            block = self.first_block()

            while block:
                # free all data blocks
                block.image.free_block(block)
                block = block.next_block()
        except ValueError:
            pass
        if self.file_type == 'REL':
            try:
                block = Block(self.block.image, *self.side_sector_ts)

                while block:
                    # free all side sectors
                    block.image.free_block(block)
                    block = block.next_block()
            except ValueError:
                pass

    def relative_file(self, mode):
        """Return relative file instance."""
        return RelativeFile(self, mode)

    @property
    def is_deleted(self):
        """Return `True` if this entry is no longer in use."""
        return self._file_type() == 0

    @property
    def file_type(self):
        return self.FTYPE_STR[self._file_type() & 7]

    @property
    def replacement(self):
        """Return `True` if file was being overwritten."""
        return bool(self._file_type() & 0x20)

    @property
    def protected(self):
        """Return `True` if file is locked from being deleted."""
        return bool(self._file_type() & 0x40)

    @property
    def closed(self):
        """Return `True` if file was closed after last write."""
        return bool(self._file_type() & 0x80)

    @property
    def start_ts(self):
        """Return track and sector of first data block."""
        return tuple(self.block.get(self.entry_offset+3, self.entry_offset+5))

    @property
    def name(self):
        """Return the name as PETSCII."""
        name = self.block.get(self.entry_offset+5, self.entry_offset+0x15)
        return name.rstrip(b'\xa0')

    @property
    def side_sector_ts(self):
        """Return track and sector of first side sector."""
        return tuple(self.block.get(self.entry_offset+0x15, self.entry_offset+0x17))

    @property
    def record_len(self):
        """Return relative file record length."""
        return self.block.get(self.entry_offset+0x17)

    @property
    def replacement_ts(self):
        """Return track and sector of new data during file replacement."""
        return tuple(self.block.get(self.entry_offset+0x1c, self.entry_offset+0x1e))

    @property
    def size(self):
        """Return file size in blocks."""
        size, = struct.unpack('<H', self.block.get(self.entry_offset+0x1e, self.entry_offset+0x20))
        return size

    @file_type.setter
    def file_type(self, ftype):
        if isinstance(ftype, str):
            if ftype.upper() not in self.FTYPE_STR:
                raise ValueError("Unknown file type, "+ftype)
            ftype = self.FTYPE_STR.index(ftype.upper()) | 0x80

        self.block.set(self.entry_offset+2, ftype)

    @replacement.setter
    def replacement(self, prot):
        val = 0x20 if prot else 0
        old = self._file_type()
        self.block.set(self.entry_offset+2, old & 0xdf | val)

    @protected.setter
    def protected(self, prot):
        val = 0x40 if prot else 0
        old = self._file_type()
        self.block.set(self.entry_offset+2, old & 0xbf | val)

    @closed.setter
    def closed(self, clsd):
        val = 0x80 if clsd else 0
        old = self._file_type()
        self.block.set(self.entry_offset+2, old & 0x7f | val)

    @start_ts.setter
    def start_ts(self, ts):
        """Modify track and sector of first data block."""
        self.block.set(self.entry_offset+3, bytes(ts))

    @name.setter
    def name(self, name):
        self.block.set(self.entry_offset+5, name[:16].ljust(16, b'\xa0'))

    @side_sector_ts.setter
    def side_sector_ts(self, ts):
        """Modify track and sector of first side sector."""
        self.block.set(self.entry_offset+0x15, bytes(ts))

    @record_len.setter
    def record_len(self, rec_len):
        """Modify relative file record length."""
        self.block.set(self.entry_offset+0x17, rec_len)

    @replacement_ts.setter
    def replacement_ts(self, ts):
        """Modify track and sector of new data during file replacement."""
        self.block.set(self.entry_offset+0x1c, bytes(ts))

    @size.setter
    def size(self, size):
        """Set file size in blocks."""
        self.block.set(self.entry_offset+0x1e, struct.pack('<H', size))

    @classmethod
    def max_file_type(cls):
        valid = [s for s in cls.FTYPE_STR if s != '???']
        return len(valid)-1



import unittest

from unittest.mock import patch, Mock

from d64.d81_bam_block import D81BAMBlock

from test.mock_block import MockBlock


class TestD81BAMBlock(unittest.TestCase):

    def setUp(self):
        mock_image = Mock()
        p = patch.object(D81BAMBlock, '__bases__', (MockBlock,))
        with p:
            p.is_local = True
            self.bam_block = D81BAMBlock(mock_image, None, None)

    def test_dos_type(self):
        p = patch.object(D81BAMBlock, '__bases__', (MockBlock,))
        with p:
            p.is_local = True
            self.bam_block.set(2, 77)
            self.assertEqual(self.bam_block.dos_type, 77)
            self.bam_block.dos_type = ord('D')
            self.assertEqual(self.bam_block.dos_type, ord('D'))

    def test_verify(self):
        p = patch.object(D81BAMBlock, '__bases__', (MockBlock,))
        with p:
            p.is_local = True
            self.bam_block.set(6, 0x80)
            self.assertTrue(self.bam_block.verify)
            self.bam_block.set(6, 0x20)
            self.assertFalse(self.bam_block.verify)
            self.bam_block.verify = True
            self.assertEqual(self.bam_block.get(6), 0xa0)

    def test_header_crc(self):
        p = patch.object(D81BAMBlock, '__bases__', (MockBlock,))
        with p:
            p.is_local = True
            self.bam_block.set(6, 0x40)
            self.assertTrue(self.bam_block.check_header_crc)
            self.bam_block.set(6, 0x20)
            self.assertFalse(self.bam_block.check_header_crc)
            self.bam_block.check_header_crc = True
            self.assertEqual(self.bam_block.get(6), 0x60)

    def test_auto_start(self):
        p = patch.object(D81BAMBlock, '__bases__', (MockBlock,))
        with p:
            p.is_local = True
            self.bam_block.set(7, 0x40)
            self.assertTrue(self.bam_block.auto_start)
            self.bam_block.set(7, 0)
            self.assertFalse(self.bam_block.auto_start)
            self.bam_block.auto_start = True
            self.assertEqual(self.bam_block.get(7), 0xff)



import sys

import test.binary

from pathlib import Path


if __name__ == '__main__':
    bin1 = test.binary.load_binary(Path(sys.argv[1]))
    bin2 = test.binary.load_binary(Path(sys.argv[2]))

    print(test.binary.diff(bin1, bin2))



from .block import Block
from .d81_bam import D81BAM
from .d81_bam_block import D81BAMBlock
from .d81_dir_entry import D81DirEntry
from .dos_image import DOSImage
from .partition import Partition


class D81Image(DOSImage):

    DIR_ENTRY = D81DirEntry
    DOS_VERSION = ord('3')
    DOS_FORMAT = ord('D')
    MAX_TRACK = 80
    DIR_TRACK = 40
    DIR_SECTOR = 3
    INTERLEAVE = 1
    SECTORS_PER_TRACK = 40
    TRACK_SECTOR_MAX = ((SECTORS_PER_TRACK, (1, 80)),)
    IMAGE_SIZES = (819200, 822400)

    def __init__(self, filename):
        self.bam = D81BAM(self)
        super().__init__(filename)

    def open(self, mode):
        super().open(mode)
        self.header_block = Block(self, self.DIR_TRACK, 0)
        self.side_a_bam_block = D81BAMBlock(self, self.DIR_TRACK, 1)
        self.side_b_bam_block = D81BAMBlock(self, self.DIR_TRACK, 2)

    def partition(self, name):
        paths = [e for e in self.glob(name)]
        if paths:
            if paths[0].entry.file_type == 'CBM':
                # existing partition
                return Partition(self, entry=paths[0].entry)
            raise TypeError("Entry is not a partition")
        return Partition(self, name=name)

    def subdirectory(self, name):
        """Return nested image representing a subdirectory."""
        # deferred import to avoid cyclic dependency
        from .subdirectory import Subdirectory

        paths = [e for e in self.glob(name)]
        if paths:
            if paths[0].entry.file_type == 'CBM':
                # existing partition
                subdir = Subdirectory(self, paths[0].entry)
                if subdir.is_formatted():
                    return subdir
            raise TypeError("Entry is not a subdirectory")
        raise FileNotFoundError("File not found: "+str(name))

    def alloc_next_block(self, track, sector, directory=False):
        return self._alloc_next_block(track, sector, self.INTERLEAVE)

    @property
    def dos_version(self):
        return self.header_block.get(0x19)

    @property
    def name(self):
        name = self.header_block.get(4, 0x14)
        return name.rstrip(b'\xa0')

    @property
    def id(self):
        id = self.header_block.get(0x16, 0x18)
        return id

    @property
    def dos_type(self):
        return self.header_block.get(2)

    @dos_version.setter
    def dos_version(self, version):
        self.header_block.set(0x19, version)

    @name.setter
    def name(self, nam):
        self.header_block.set(4, nam[:16].ljust(16, b'\xa0'))

    @id.setter
    def id(self, did):
        if len(did) != 2:
            raise ValueError("Invalid disk id, "+str(did))
        self.header_block.set(0x16, did)
        # update copy in BAM blocks
        self.side_a_bam_block.id = did
        self.side_b_bam_block.id = did

    @dos_type.setter
    def dos_type(self, dtype):
        self.header_block.set(2, dtype)
        self.header_block.set(0x1a, dtype)
        # update copy in BAM blocks
        self.side_a_bam_block.dos_type = dtype
        self.side_b_bam_block.dos_type = dtype

    def _create(self, disk_name, disk_id):
        """Populate an image or subdirectory."""
        # block 40/0 contains various identifying fields
        header_block = Block(self, self.DIR_TRACK, 0)
        header_block.set(4, b'\xa0' * 0x15)
        self.name = disk_name
        self.id = disk_id
        self.dos_type = self.DOS_FORMAT
        self.dos_version = self.DOS_VERSION

        # populate the BAM with all free blocks
        for track in range(self.MIN_TRACK, self.MAX_TRACK+1):
            bits_free = '1' * self.SECTORS_PER_TRACK
            self.bam.set_entry(track, self.SECTORS_PER_TRACK, bits_free)

        # block 40/3 contains 8 empty directory entries
        dir_block = Block(self, self.DIR_TRACK, self.DIR_SECTOR)
        dir_block.data_size = 0xfe

        # blocks 40/1 and 40/2 hold the BAM
        bam1_block = D81BAMBlock(self, self.DIR_TRACK, 1)
        bam2_block = D81BAMBlock(self, self.DIR_TRACK, 2)
        bam2_block.data_size = 0xfe

        for block in (bam1_block, bam2_block):
            block.id = disk_id
            block.dos_type = self.DOS_FORMAT
            block.verify = True
            block.check_header_crc = True

        # link BAM 1 block to BAM 2 block
        bam1_block.set_next_block(bam2_block)

        # link header block to directory block
        header_block.set_next_block(dir_block)

        # allocate all blocks
        for i in range(0, 4):
            self.bam.set_allocated(self.DIR_TRACK, i)

    @classmethod
    def create(cls, filepath, disk_name, disk_id):
        """Create an empty disk image."""
        super().create(filepath)

        image = cls(filepath)
        try:
            image.open('r+b')
            image._create(disk_name, disk_id)
        finally:
            image.close()



import logging

from .exceptions import ConsistencyError, DiskFullError
from .path_base import PathBase

log = logging.getLogger(__name__)


class Partition(PathBase):
    """Partition, unstructured sequence of blocks (1581)."""
    def create(self, block_start, size_blocks):
        """Create a new partition."""
        if self.entry:
            raise FileExistsError("File exists: "+str(self.name))

        log.debug("Creating partition %s, start %d:%d, size %d blocks", self.name, block_start.track, block_start.sector, size_blocks)
        entry = self.image.get_free_entry()
        if entry is None:
            raise DiskFullError()

        entry.file_type = 'CBM'
        entry.name = self._name
        entry.size = size_blocks
        entry.set_first_block(block_start)

        # allocate all blocks
        alloc_blocks = []
        try:
            for block in entry.partition_blocks():
                self.image.bam.set_allocated(block.track, block.sector)
                alloc_blocks.append(block)
        except (ValueError, ConsistencyError):
            # free blocks that were allocated
            for block in alloc_blocks:
                self.image.free_block(block)
            entry.file_type = 0
            entry.reset()
            raise

        self.entry = entry
        self._name = None

    def format(self, disk_name, disk_id):
        from .subdirectory import Subdirectory

        log.debug("Format partition %s as %s:%s", self.entry.name, disk_name, disk_id)
        Subdirectory.create(self.image, self.entry, disk_name, disk_id)

    def blocks(self):
        """Generator to return each block of a partition."""
        self.assert_has_entry()

        yield from self.entry.partition_blocks()



import json


def load_binary(path):
    """Return the contents of a file as a byte string."""
    with path.open('rb') as fileh:
        return fileh.read()


def load_patch(path):
    """Return a patch as an array of changes."""
    with path.open() as fileh:
        return json.load(fileh)


def diff(bin1, bin2):
    """Return the difference between two byte strings as an array of changes."""
    if len(bin1) != len(bin2):
        raise ValueError("Binary lengths differ")

    ret = []
    current_delta = None

    for i in range(0, len(bin1)):
        if bin1[i] == bin2[i]:
            if current_delta:
                ret.append(current_delta)
                current_delta = None
        else:
            if current_delta is None:
                current_delta = {'at': i, 'from': bytes([bin1[i]]), 'to': bytes([bin2[i]])}
            else:
                current_delta['from'] += bytes([bin1[i]])
                current_delta['to'] += bytes([bin2[i]])

    if current_delta:
        ret.append(current_delta)

    return ret


def patch(bin, patch, out_path):
    """Create a new file by patching a byte string."""
    pos = 0

    with out_path.open('wb') as fileh:
        for delta in patch:
            if pos != delta['at']:
                fileh.write(bin[pos:delta['at']])
                pos = delta['at']
            fileh.write(delta['to'])
            pos += len(delta['to'])

        if pos != len(bin):
            fileh.write(bin[pos:])



import unittest
from unittest.mock import Mock

from d64.d71_bam import D71BAM

from test.mock_block import MockBlock


class TestBAM(unittest.TestCase):

    def setUp(self):
        self.image = Mock()
        self.image.MIN_TRACK = 1
        self.image.MAX_TRACK = 70
        self.image.bam_block = MockBlock()
        self.image.extra_bam_block = MockBlock()
        self.image.bam = D71BAM(self.image)

    def test_get_entry(self):
        self.image.bam_block.data[12:16] = b'\x0d\xc9\x6e\x1b'
        entry = self.image.bam.get_entry(3)
        self.assertEqual(entry, (13, '100100110111011011011000'))
        self.image.bam_block.data[0xe8] = 0x0d
        self.image.extra_bam_block.data[33:36] = b'\xc9\x6e\x1b'
        entry = self.image.bam.get_entry(47)
        self.assertEqual(entry, (13, '100100110111011011011000'))

    def test_set_entry(self):
        self.image.bam.set_entry(3, 15, '110111011100111101011000')
        self.assertEqual(self.image.bam_block.data[12:16], b'\x0f\xbb\xf3\x1a')
        self.image.bam.set_entry(47, 15, '110111011100111101011000')
        self.assertEqual(self.image.bam_block.data[0xe8], 0x0f)
        self.assertEqual(self.image.extra_bam_block.data[33:36], b'\xbb\xf3\x1a')


if __name__ == '__main__':
    unittest.main()



from .block import Block
from .d80_d82_bam import D80D82BAM
from .d80_d82_bam_block import D80D82BAMBlock
from .dos_image import DOSImage


class D80Image(DOSImage):

    DOS_VERSION = ord('2')
    DOS_FORMAT = ord('C')
    MAX_TRACK = 77
    BAM_TRACK = 38
    BAM_SECTORS = (0, 3)
    BAM_TRACK_RANGES = ((1, 50), (51, 77))
    DIR_TRACK = 39
    DIR_SECTOR = 1
    INTERLEAVE = 1
    TRACK_SECTOR_MAX = ((29, (1, 39)), (27, (40, 53)), (25, (54, 64)), (23, (65, 77)))
    IMAGE_SIZES = (533248, )

    def __init__(self, filename):
        super().__init__(filename)
        self.bam = D80D82BAM(self)

    def open(self, mode):
        super().open(mode)
        self.header_block = Block(self, self.DIR_TRACK, 0)
        self.bam_blocks = [D80D82BAMBlock(self, self.BAM_TRACK, s) for s in self.BAM_SECTORS]

    def alloc_next_block(self, track, sector, directory=False):
        return self._alloc_next_block(track, sector, self.INTERLEAVE)

    @property
    def dos_version(self):
        return self.header_block.get(0x1b)

    @property
    def name(self):
        name = self.header_block.get(0x06, 0x16)
        return name.rstrip(b'\xa0')

    @property
    def id(self):
        id = self.header_block.get(0x18, 0x1a)
        return id

    @property
    def dos_type(self):
        return self.header_block.get(2)

    @dos_version.setter
    def dos_version(self, version):
        self.header_block.set(0x1b, version)

    @name.setter
    def name(self, nam):
        self.header_block.set(6, nam[:16].ljust(16, b'\xa0'))

    @id.setter
    def id(self, did):
        if len(did) != 2:
            raise ValueError("Invalid disk id, "+str(did))
        self.header_block.set(0x18, did)

    @dos_type.setter
    def dos_type(self, dtype):
        self.header_block.set(2, dtype)
        self.header_block.set(0x1c, dtype)

    @classmethod
    def create(cls, filepath, disk_name, disk_id):
        """Create an empty disk image."""
        super().create(filepath)

        image = cls(filepath)
        try:
            image.open('r+b')

            # block 39/0 contains various identifying fields
            header_block = Block(image, cls.DIR_TRACK, 0)
            header_block.set(6, b'\xa0' * 0x1b)
            image.name = disk_name
            image.id = disk_id
            image.dos_type = cls.DOS_FORMAT
            image.dos_version = cls.DOS_VERSION

            # BAM spans multiple blocks on track 38
            prev_block = header_block
            bam_blocks = []
            for sector, track_range in zip(cls.BAM_SECTORS, cls.BAM_TRACK_RANGES):
                bam_block = D80D82BAMBlock(image, cls.BAM_TRACK, sector)
                bam_block.dos_type = cls.DOS_FORMAT
                bam_block.track_range = (track_range[0], track_range[1]+1)
                # link BAM blocks
                prev_block.set_next_block(bam_block)
                prev_block = bam_block
                bam_blocks.append(bam_block)

            # populate the BAM with all free blocks
            for sectors, range_ in cls.TRACK_SECTOR_MAX:
                for t in range(range_[0], range_[1]+1):
                    bits_free = '1' * sectors
                    image.bam.set_entry(t, sectors, bits_free)

            # block 39/1 contains 8 empty directory entries
            dir_block = Block(image, cls.DIR_TRACK, cls.DIR_SECTOR)
            dir_block.data_size = 0xfe

            # link final BAM block to directory block
            bam_blocks[-1].set_next_block(dir_block)

            # allocate all blocks
            for block in [header_block, dir_block]+bam_blocks:
                image.bam.set_allocated(block.track, block.sector)
        finally:
            image.close()



import unittest

from contextlib import suppress
from pathlib import Path

import d64.scripts.d64_fsck

from d64.d80_image import D80Image

import binary


class TestD80_fsck(unittest.TestCase):

    def setUp(self):
        self.base_path = Path(__file__).parent / 'data' / 'test.d80'
        self.test_path = Path('/tmp/test_bad.d80')
        self.base_bin = binary.load_binary(self.base_path)
        d64.scripts.d64_fsck.QUIET = True
        d64.scripts.d64_fsck.FIX = True
        d64.scripts.d64_fsck.YES = True

    def tearDown(self):
        with suppress(FileNotFoundError):
            self.test_path.unlink()

    def test_clean(self):
        d64.scripts.d64_fsck.FIX = False
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.base_path), 0)

    def test_bad_dir_link(self):
        patch = [{'at': 282113, 'from': b'\x00', 'to': b'\x01'}]
        binary.patch(self.base_bin, patch, self.test_path)
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.test_path), 0)
        image = D80Image(self.test_path)
        try:
            image.open('rb')
            self.assertEqual(image.header_block.next_block().sector, 0)
        finally:
            image.close()

    def test_bam_38_03_not_alloc(self):
        patch = [{'at': 274884, 'from': b'\x1b\xfc', 'to': b'\x1c\xfe'}]
        binary.patch(self.base_bin, patch, self.test_path)
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.test_path), 0)
        image = D80Image(self.test_path)
        try:
            image.open('rb')
            self.assertEqual(image.bam.get_entry(38), (27, '01101111111111111111111111111000'))
        finally:
            image.close()


if __name__ == '__main__':
    unittest.main()



import unittest

from unittest.mock import patch, Mock

from d64.d81_relative_file import D81RelativeFile
from d64.side_sector import SideSector
from d64.super_side_sector import SuperSideSector

from test.mock_block import MockBlock


class TestD81RelFile(unittest.TestCase):

    def setUp(self):
        mock_image = Mock()
        mock_image.alloc_first_block.return_value = MockBlock(mock_image)
        mock_image.alloc_next_block.side_effect = lambda x, y: MockBlock(mock_image)
        self.entry = Mock()
        self.entry.size = 0
        self.entry.block = MockBlock(mock_image)
        self.file = D81RelativeFile(self.entry, 'w')
        self.file.image = mock_image

    def test_write_short(self):
        self.entry.record_len = 27
        p1 = patch.object(SideSector, '__bases__', (MockBlock,))
        with p1:
            p1.is_local = True
            p2 = patch.object(SuperSideSector, '__bases__', (MockBlock,))
            with p2:
                p2.is_local = True
                self.assertEqual(self.file.write(b'abcdefg\r'), 27)
                self.assertEqual(self.file.side_sector.number, 0)
                self.assertEqual(self.file.side_sector.record_len, 27)
                self.assertEqual(len(self.file.side_sector.all_side_sectors()), 1)
                self.assertEqual(len(self.file.side_sector.all_data_blocks()), 1)
        self.assertEqual(self.file.block.data_size, 27)
        self.assertEqual(self.file.block.data[2:29], b'abcdefg\r'+b'\x00'*19)
        self.assertEqual(self.entry.size, 3)



import unittest

from d64 import DOSPath


class TestWildcards(unittest.TestCase):

    def test_wildcards_plain(self):
        self.assertTrue(DOSPath.wildcard_match(b'EXACT', 'PRG', b'EXACT'))
        self.assertFalse(DOSPath.wildcard_match(b'EXYCT', 'PRG', b'EXACT'))
        self.assertFalse(DOSPath.wildcard_match(b'EXACTLY', 'PRG', b'EXACT'))
        self.assertFalse(DOSPath.wildcard_match(b'EXA', 'PRG', b'EXACT'))

    def test_wildcards_single(self):
        self.assertTrue(DOSPath.wildcard_match(b'SINGLE', 'PRG', b'SINGL?'))
        self.assertTrue(DOSPath.wildcard_match(b'SINGLE', 'PRG', b'?INGLE'))
        self.assertFalse(DOSPath.wildcard_match(b'SINGLE', 'PRG', b'SINGLE?'))

    def test_wildcards_multi(self):
        self.assertTrue(DOSPath.wildcard_match(b'MULTIPLE', 'PRG', b'MULTIPLE*'))
        self.assertTrue(DOSPath.wildcard_match(b'MULTIPLE', 'PRG', b'MULTIPL*'))
        self.assertTrue(DOSPath.wildcard_match(b'MULTIPLE', 'PRG', b'MULTIP*'))
        self.assertTrue(DOSPath.wildcard_match(b'MULTIPLE', 'PRG', b'*'))

    def test_wildcards_type(self):
        self.assertTrue(DOSPath.wildcard_match(b'TYPE', 'PRG', b'*=PRG'))
        self.assertTrue(DOSPath.wildcard_match(b'TYPE', 'PRG', b'*=P'))
        self.assertFalse(DOSPath.wildcard_match(b'TYPE', 'PRG', b'*=S'))

    def test_wildcards_mixed(self):
        self.assertTrue(DOSPath.wildcard_match(b'TYPE', 'PRG', b'TYPE=PRG'))
        self.assertFalse(DOSPath.wildcard_match(b'TYPER', 'PRG', b'TYPE=PRG'))
        self.assertFalse(DOSPath.wildcard_match(b'TYPE', 'PRG', b'TYPE=SEQ'))
        self.assertTrue(DOSPath.wildcard_match(b'TYPE', 'PRG', b'TYPE*=PRG'))
        self.assertFalse(DOSPath.wildcard_match(b'TYPE', 'PRG', b'TYPE*=SEQ'))

    def test_wildcards_bad(self):
        with self.assertRaises(ValueError):
            DOSPath.wildcard_match(b'EXACT', 'PRG', b'EXACT=')


if __name__ == '__main__':
    unittest.main()




class PathBase(object):
    def __init__(self, image, entry=None, name=None):
        self.image = image
        self.entry = entry
        self._name = name

    def exists(self):
        """Return `True` if path exists."""
        return self.entry is not None

    def assert_has_entry(self):
        if self.entry is None:
            raise FileNotFoundError("File not found: "+str(self.name))

    def unlink(self):
        """Delete directory entry for this path."""
        self.assert_has_entry()

        self.entry.free_blocks()
        self.entry.file_type = 0

    def rename(self, other):
        """Rename this directory entry, replacing another entry if it exists."""
        self.assert_has_entry()

        if isinstance(other, str) or isinstance(other, bytes):
            other = self.image.path(other)

        if other.entry:
            other.unlink()
        self.entry.name = other.name
        return self

    replace = rename

    @property
    def name(self):
        if self.entry:
            return self.entry.name
        return self._name

    @property
    def size_blocks(self):
        """Return file size in blocks."""
        self.assert_has_entry()

        return self.entry.size



import unittest

from contextlib import suppress
from pathlib import Path

import d64.scripts.d64_fsck

from d64.block import Block
from d64.d64_image import D64Image
from d64.side_sector import SideSector

import binary


class TestD64_fsck(unittest.TestCase):

    def setUp(self):
        self.base_path = Path(__file__).parent / 'data' / 'test.d64'
        self.test_path = Path('/tmp/test_bad.d64')
        self.base_bin = binary.load_binary(self.base_path)
        d64.scripts.d64_fsck.QUIET = True
        d64.scripts.d64_fsck.FIX = True
        d64.scripts.d64_fsck.YES = True

    def tearDown(self):
        with suppress(FileNotFoundError):
            self.test_path.unlink()

    def test_clean(self):
        d64.scripts.d64_fsck.FIX = False
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.base_path), 0)

    def test_dos_version(self):
        patch = [{'at': 91557, 'from': b'2', 'to': b'7'}]
        binary.patch(self.base_bin, patch, self.test_path)
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.test_path), 0)
        image = D64Image(self.test_path)
        try:
            image.open('rb')
            self.assertEqual(image.dos_version, ord('2'))
        finally:
            image.close()

    def test_dos_format(self):
        patch = [{'at': 91558, 'from': b'A', 'to': b'C'}]
        binary.patch(self.base_bin, patch, self.test_path)
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.test_path), 0)
        image = D64Image(self.test_path)
        try:
            image.open('rb')
            self.assertEqual(image.dos_type, ord('A'))
        finally:
            image.close()

    def test_dir_link(self):
        patch = [{'at': 91392, 'from': b'\x12', 'to': b'\x00'}]
        binary.patch(self.base_bin, patch, self.test_path)
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.test_path), 0)
        image = D64Image(self.test_path)
        try:
            image.open('rb')
            self.assertEqual(repr(image.bam_block), '<Block 18:0 (18:1)>')
        finally:
            image.close()

    def test_bam_entry_diff(self):
        patch = [{'at': 91416, 'from': b'\x15', 'to': b'\x16'}]
        binary.patch(self.base_bin, patch, self.test_path)
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.test_path), 0)
        image = D64Image(self.test_path)
        try:
            image.open('rb')
            self.assertEqual(image.bam.get_entry(6), (21, '111111111111111111111000'))
        finally:
            image.close()

    def test_bam_18_00_not_alloc(self):
        patch = [{'at': 91464, 'from': b'\x0el', 'to': b'\x0fm'}]
        binary.patch(self.base_bin, patch, self.test_path)
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.test_path), 0)
        image = D64Image(self.test_path)
        try:
            image.open('rb')
            self.assertEqual(image.bam.get_entry(18), (14, '001101101101111111100000'))
        finally:
            image.close()

    def test_dir_18_01_not_alloc(self):
        patch = [{'at': 91464, 'from': b'\x0el', 'to': b'\x0fn'}]
        binary.patch(self.base_bin, patch, self.test_path)
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.test_path), 0)
        image = D64Image(self.test_path)
        try:
            image.open('rb')
            self.assertEqual(image.bam.get_entry(18), (14, '001101101101111111100000'))
        finally:
            image.close()

    def test_dir_loop(self):
        patch = [{'at': 93185, 'from': b'\n', 'to': b'\x01'}]
        binary.patch(self.base_bin, patch, self.test_path)
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.test_path), 0)
        image = D64Image(self.test_path)
        try:
            image.open('rb')
            self.assertEqual(len([e for e in image.iterdir()]), 24)
        finally:
            image.close()

    def test_dir_bad_link(self):
        patch = [{'at': 93185, 'from': b'\n', 'to': b'\xc7'}]
        binary.patch(self.base_bin, patch, self.test_path)
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.test_path), 0)
        image = D64Image(self.test_path)
        try:
            image.open('rb')
            self.assertEqual(len([e for e in image.iterdir()]), 24)
        finally:
            image.close()

    def test_dir_end_len(self):
        patch = [{'at': 93953, 'from': b'\xff', 'to': b'\x87'}]
        binary.patch(self.base_bin, patch, self.test_path)
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.test_path), 0)
        image = D64Image(self.test_path)
        try:
            image.open('rb')
            self.assertEqual(Block(image, 18, 10).data_size, 0xfe)
        finally:
            image.close()

    def test_ent_bad_type(self):
        patch = [{'at': 92482, 'from': b'\x82', 'to': b'\x87'}]
        binary.patch(self.base_bin, patch, self.test_path)
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.test_path), 0)
        image = D64Image(self.test_path)
        try:
            image.open('rb')
            self.assertEqual(image.path(b'SMALL10').entry.file_type, 'PRG')
        finally:
            image.close()

    def test_ent_bad_1st(self):
        patch = [{'at': 92483, 'from': b'\x11', 'to': b'I'}]
        binary.patch(self.base_bin, patch, self.test_path)
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.test_path), 0)
        image = D64Image(self.test_path)
        try:
            image.open('rb')
            self.assertFalse(image.path(b'SMALL10').exists())
        finally:
            image.close()

    def test_file_replacement(self):
        patch = [{'at': 91650, 'from': b'\x81', 'to': b'!'},
                 {'at': 91676, 'from': b'\x00\x00', 'to': b'\x11\x01'},
                 {'at': 91682, 'from': b'\x81\x11\x01', 'to': b'\x00\x00\x00'}]
        binary.patch(self.base_bin, patch, self.test_path)
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.test_path), 0)
        image = D64Image(self.test_path)
        try:
            image.open('rb')
            self.assertEqual(image.path(b'SMALL1').size_bytes, 1000)
        finally:
            image.close()

    def test_file_unclosed(self):
        patch = [{'at': 92482, 'from': b'\x82', 'to': b'\x02'}]
        binary.patch(self.base_bin, patch, self.test_path)
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.test_path), 0)
        image = D64Image(self.test_path)
        try:
            image.open('rb')
            self.assertTrue(image.path(b'SMALL10').entry.closed)
        finally:
            image.close()

    def test_file_unalloc(self):
        patch = [{'at': 91468, 'from': b'\x00\x00', 'to': b'\x01 '}]
        binary.patch(self.base_bin, patch, self.test_path)
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.test_path), 0)
        image = D64Image(self.test_path)
        try:
            image.open('rb')
            self.assertEqual(image.bam.get_entry(19), (0, '000000000000000000000000'))
        finally:
            image.close()

    def test_file_loop_1st(self):
        patch = [{'at': 93284, 'from': b'\x04', 'to': b'\x02'}]
        binary.patch(self.base_bin, patch, self.test_path)
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.test_path), 0)
        image = D64Image(self.test_path)
        try:
            image.open('rb')
            self.assertNotEqual(image.path(b'LARGE4').entry.first_block, image.path(b'LARGE5').entry.first_block)
        finally:
            image.close()

    def test_file_bad_link(self):
        patch = [{'at': 84481, 'from': b'\x06', 'to': b'`'}]
        binary.patch(self.base_bin, patch, self.test_path)
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.test_path), 0)
        image = D64Image(self.test_path)
        try:
            image.open('rb')
            self.assertEqual(image.path(b'LARGE10').size_blocks, 2)
        finally:
            image.close()

    def test_file_loop(self):
        patch = [{'at': 84481, 'from': b'\x06', 'to': b'\x05'}]
        binary.patch(self.base_bin, patch, self.test_path)
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.test_path), 0)
        image = D64Image(self.test_path)
        try:
            image.open('rb')
            self.assertEqual(image.path(b'LARGE10').size_blocks, 2)
        finally:
            image.close()

    def test_file_xlink(self):
        patch = [{'at': 99841, 'from': b'\x05', 'to': b'\x03'}]
        binary.patch(self.base_bin, patch, self.test_path)
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.test_path), 0)
        image = D64Image(self.test_path)
        try:
            image.open('rb')
            self.assertNotEqual(Block(image, 19, 14).next_block(), Block(image, 19, 3))
            self.assertEqual(image.path(b'LARGE10').size_blocks, 4)
        finally:
            image.close()

    def test_file_bad_len(self):
        patch = [{'at': 87553, 'from': b'e', 'to': b'\x00'}]
        binary.patch(self.base_bin, patch, self.test_path)
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.test_path), 0)
        image = D64Image(self.test_path)
        try:
            image.open('rb')
            self.assertEqual(image.path(b'SMALL5').size_bytes, 254)
        finally:
            image.close()

    def test_ss_bad_index(self):
        patch = [{'at': 137986, 'from': b'\x01', 'to': b'\x05'}]
        binary.patch(self.base_bin, patch, self.test_path)
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.test_path), 0)
        image = D64Image(self.test_path)
        try:
            image.open('rb')
            side_sector = SideSector(image, *image.path(b'DATABASE').entry.side_sector_ts)
            self.assertEqual(side_sector.next_block().number, 1)
        finally:
            image.close()

    def test_ss_bad_rec_len(self):
        patch = [{'at': 137987, 'from': b'\xc8', 'to': b'\x8c'}]
        binary.patch(self.base_bin, patch, self.test_path)
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.test_path), 0)
        image = D64Image(self.test_path)
        try:
            image.open('rb')
            side_sector = SideSector(image, *image.path(b'DATABASE').entry.side_sector_ts)
            self.assertEqual(side_sector.next_block().record_len, 200)
        finally:
            image.close()

    def test_ss_unalloc(self):
        patch = [{'at': 91500, 'from': b'\x00', 'to': b'\x01'}, {'at': 91502, 'from': b'\x00', 'to': b' '}]
        binary.patch(self.base_bin, patch, self.test_path)
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.test_path), 0)
        image = D64Image(self.test_path)
        try:
            image.open('rb')
            self.assertEqual(image.bam.get_entry(27), (0, '000000000000000000000000'))
        finally:
            image.close()

    def test_ss_peer_ss(self):
        patch = [{'at': 137988, 'from': b'\x14', 'to': b'w'}]
        binary.patch(self.base_bin, patch, self.test_path)
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.test_path), 0)
        image = D64Image(self.test_path)
        try:
            image.open('rb')
            side_sector = SideSector(image, *image.path(b'DATABASE').entry.side_sector_ts)
            self.assertEqual(side_sector.next_block().all_side_sectors(), [(20, 18), (27, 13), (34, 14)])
        finally:
            image.close()

    def test_ss_data_link(self):
        patch = [{'at': 138122, 'from': b'\x1e', 'to': b'-'}]
        binary.patch(self.base_bin, patch, self.test_path)
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.test_path), 0)
        image = D64Image(self.test_path)
        try:
            image.open('rb')
            side_sector = SideSector(image, *image.path(b'DATABASE').entry.side_sector_ts)
            self.assertEqual(side_sector.next_block().all_data_blocks()[61], (30, 6))
        finally:
            image.close()

    def test_rel_bad_1st_ss(self):
        patch = [{'at': 94198, 'from': b'\x12', 'to': b'\x97'}]
        binary.patch(self.base_bin, patch, self.test_path)
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.test_path), 0)
        image = D64Image(self.test_path)
        try:
            image.open('rb')
            side_sector = SideSector(image, *image.path(b'DATABASE').entry.side_sector_ts)
            self.assertEqual(len(side_sector.all_side_sectors()), 3)
            self.assertEqual(side_sector.record_len, 200)
        finally:
            image.close()

    def test_rel_xlink_1st_ss(self):
        patch = [{'at': 94197, 'from': b'\x14\x12', 'to': b'\x12\x01'}]
        binary.patch(self.base_bin, patch, self.test_path)
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.test_path), 0)
        image = D64Image(self.test_path)
        try:
            image.open('rb')
            side_sector = SideSector(image, *image.path(b'DATABASE').entry.side_sector_ts)
            self.assertEqual(len(side_sector.all_side_sectors()), 3)
            self.assertEqual(side_sector.record_len, 200)
        finally:
            image.close()

    def test_rel_bad_ss(self):
        patch = [{'at': 137985, 'from': b'\x0e', 'to': b'\x9a'}]
        binary.patch(self.base_bin, patch, self.test_path)
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.test_path), 0)
        image = D64Image(self.test_path)
        try:
            image.open('rb')
            side_sector = SideSector(image, *image.path(b'DATABASE').entry.side_sector_ts)
            self.assertEqual(len(side_sector.all_side_sectors()), 3)
            self.assertEqual(side_sector.record_len, 200)
        finally:
            image.close()

    def test_rel_loop_ss(self):
        patch = [{'at': 137984, 'from': b'"\x0e', 'to': b'\x14\x12'}]
        binary.patch(self.base_bin, patch, self.test_path)
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.test_path), 0)
        image = D64Image(self.test_path)
        try:
            image.open('rb')
            side_sector = SideSector(image, *image.path(b'DATABASE').entry.side_sector_ts)
            self.assertEqual(len(side_sector.all_side_sectors()), 3)
            self.assertEqual(side_sector.record_len, 200)
        finally:
            image.close()

    def test_rel_xlink_ss(self):
        patch = [{'at': 137984, 'from': b'"\x0e', 'to': b'\x12\x01'}]
        binary.patch(self.base_bin, patch, self.test_path)
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.test_path), 0)
        image = D64Image(self.test_path)
        try:
            image.open('rb')
            side_sector = SideSector(image, *image.path(b'DATABASE').entry.side_sector_ts)
            self.assertEqual(len(side_sector.all_side_sectors()), 3)
            self.assertEqual(side_sector.record_len, 200)
        finally:
            image.close()

    def test_rel_too_many_ss(self):
        patch = [{'at': 75520, 'from': b'\x00\x00', 'to': b'\x0f\x02'}, {'at': 75776, 'from': b'\x00\x00', 'to': b'\x0f\x03'},
                 {'at': 76032, 'from': b'\x00\x00', 'to': b'\x0f\x04'}, {'at': 91452, 'from': b'\x0f\xbe', 'to': b'\x0b\xa0'},
                 {'at': 169728, 'from': b'\x00W', 'to': b'\x0f\x01'}]
        binary.patch(self.base_bin, patch, self.test_path)
        self.assertEqual(d64.scripts.d64_fsck.check_image(self.test_path), 0)
        image = D64Image(self.test_path)
        try:
            image.open('rb')
            side_sector = SideSector(image, *image.path(b'DATABASE').entry.side_sector_ts)
            self.assertEqual(len(side_sector.all_side_sectors()), 3)
        finally:
            image.close()


if __name__ == '__main__':
    unittest.main()



import logging

from .exceptions import DiskFullError, FileIndexError
from .file import File
from .relative_file import RelativeFile
from .side_sector import SideSector
from .super_side_sector import SuperSideSector

log = logging.getLogger(__name__)


class D81RelativeFile(RelativeFile):
    def get_first_block(self):
        """Allocate side sector and initialize it."""
        data_block = File.get_first_block(self)
        side_sector = self.get_first_side_sector(data_block)
        if side_sector is None:
            self.image.free_block(data_block)
            raise DiskFullError()

        block = self.image.alloc_next_block(side_sector.track, side_sector.sector)
        if block is None:
            self.image.free_block(data_block)
            self.image.free_block(side_sector)
            raise DiskFullError()

        side_sector.add_data_block(data_block)
        self.side_sector = side_sector
        super_side_sector = SuperSideSector(self.image, block.track, block.sector)
        log.debug("Allocated super side sector %s", super_side_sector)
        super_side_sector.clear_side_sectors()
        super_side_sector.set_next_block(side_sector)
        super_side_sector.add_side_sector(side_sector)
        self.entry.side_sector_ts = (super_side_sector.track, super_side_sector.sector)
        self.entry.size += 2
        return data_block

    def get_next_side_sector(self, data_block):
        """Allocate the next side sector."""
        if self.side_sector.number+1 == SideSector.MAX_SIDE_SECTORS:
            # current side sector group is full, create a new one
            side_sector = self.get_first_side_sector(data_block)
            if side_sector is None:
                raise DiskFullError()

            super_side_sector = SuperSideSector(self.image, *self.entry.side_sector_ts)
            if super_side_sector.add_side_sector(side_sector) == -1:
                # super side sector is full
                self.image.free_block(side_sector)
                raise FileIndexError()

            self.side_sector = side_sector
            self.entry.size += 1
            return

        super().get_next_side_sector(data_block)



from .block import Block
from .side_sector import SideSector


class SuperSideSector(Block):
    """Relative file super side sector block (1581)."""
    MAX_SIDE_SECTORS = 126
    SUPER_ID = 0xfe

    @property
    def super_id(self):
        """Return super side sector identifier."""
        return self.get(2)

    def set_super_id(self):
        """Set super side sector identifier."""
        self.set(2, self.SUPER_ID)

    def all_side_sectors(self):
        """Return an array of all side sectors track & sector."""
        ss_bin = self.get(3, 0xff)
        return [(ss_bin[i], ss_bin[i+1]) for i in range(0, self.MAX_SIDE_SECTORS*2, 2) if ss_bin[i]]

    def side_sector(self, idx):
        """Return a given side sector."""
        if idx >= self.MAX_SIDE_SECTORS:
            raise ValueError("Invalid side sector index, %d" % idx)
        return SideSector(self.image, self.get(3+idx*2), self.get(4+idx*2))

    def set_side_sector(self, idx, side_sector):
        """Set side sector in side sector list."""
        if idx >= self.MAX_SIDE_SECTORS:
            raise ValueError("Invalid side sector index, %d" % idx)
        self.set(3+idx*2, bytes((side_sector.track, side_sector.sector)))

    def clear_side_sectors(self):
        """Zero out all side sector links."""
        self.set_super_id()
        self.set(3, bytes(0xfd))

    def add_side_sector(self, side_sector):
        """Append a new side sector to the side sector list."""
        for idx in range(0, self.MAX_SIDE_SECTORS*2, 2):
            if self.get(idx+3) == 0:
                self.set(idx+3, bytes((side_sector.track, side_sector.sector)))
                return idx

        return -1

    def next_block(self):
        """Return side sector linked from the super side sector."""
        if self.is_final:
            return None
        return SideSector(self.image, self.get(0), self.get(1))



import sys

import petscii_codecs

from cbm_files import ProgramFile
from d64 import DiskImage


with DiskImage(sys.argv[1]) as image:
    with image.path(sys.argv[2].encode('petscii-c64en-uc')).open() as f:
        p = ProgramFile(f)

for line in p.to_text():
    print(line)



class MockBlock(object):
    SECTOR_SIZE = 256
    BLOCK_FILL = bytes(64) * 4

    def __init__(self, image=None, track=1, sector=0):
        self.image = image
        self.track = track
        self.sector = sector
        self.data = bytearray(self.BLOCK_FILL)
        self._data_size = 0
        self.start = 0
        self.is_final = True
        self._next_block = None

    def get(self, start, end=None):
        if end is None:
            return self.data[start]
        return self.data[start:end]

    def set(self, start, new):
        if isinstance(new, int):
            self.data[start] = new
        else:
            self.data[start:start+len(new)] = new

    def _set_data(self, data):
        self.data = data
        self.data_size = len(data)-2
        self._next_block = None
        self.is_final = True

    def set_next_block(self, block):
        self.data_size = 254
        self._next_block = block
        self.is_final = False

    def next_block(self):
        return self._next_block

    @property
    def data_size(self):
        return self._data_size

    @data_size.setter
    def data_size(self, size):
        self._data_size = size
        self._next_block = None
        self.is_final = True

    def __repr__(self):
        if self._next_block:
            t = self._next_block.track
            s = self._next_block.sector
        else:
            t = 0
            s = self.data_size
        return "<MockBlock {}:{} ({}:{})>".format(self.track, self.sector, t, s)



import unittest

from d64.d71_image import D71Image


class TestD71ImageBlocks(unittest.TestCase):

    def setUp(self):
        self.image = D71Image(None)

    def test_max_sectors(self):
        self.assertEqual(self.image.max_sectors(1), 21)
        self.assertEqual(self.image.max_sectors(5), 21)
        self.assertEqual(self.image.max_sectors(20), 19)
        self.assertEqual(self.image.max_sectors(30), 18)
        self.assertEqual(self.image.max_sectors(50), 21)
        self.assertEqual(self.image.max_sectors(55), 19)
        self.assertEqual(self.image.max_sectors(65), 18)
        self.assertEqual(self.image.max_sectors(70), 17)

    def test_max_sectors_bad(self):
        with self.assertRaises(ValueError):
            self.image.max_sectors(0)
        with self.assertRaises(ValueError):
            self.image.max_sectors(71)

    def test_block_start(self):
        self.assertEqual(self.image.block_start(1, 0), 0)
        self.assertEqual(self.image.block_start(1, 1), 256)
        self.assertEqual(self.image.block_start(1, 20), 5120)
        self.assertEqual(self.image.block_start(2, 0), 5376)
        self.assertEqual(self.image.block_start(35, 16), 174592)

    def test_block_start_bad(self):
        with self.assertRaises(ValueError):
            self.image.block_start(0, 0)
        with self.assertRaises(ValueError):
            self.image.block_start(36, 21)


if __name__ == '__main__':
    unittest.main()



from d64.bam import BAM


class MockBAM(BAM):
    def __init__(self, track_sector_max):
        self.entries = [None]
        self.max_sectors = [None]
        self.allocated = set()
        for sectors, range_ in track_sector_max:
            for t in range(range_[0], range_[1]+1):
                bits_free = '1' * sectors
                self.entries.append((sectors, bits_free))
                self.max_sectors.append(sectors)

    def fill_entry(self, track):
        _, free_bits = self.entries[track]
        self.entries[track] = (self.max_sectors[track], free_bits.replace('0', '1'))

    def clear_entry(self, track):
        _, free_bits = self.entries[track]
        self.entries[track] = (0, free_bits.replace('1', '0'))

    def get_entry(self, track):
        return self.entries[track]

    def set_allocated(self, track, sector):
        total, free_bits = self.entries[track]
        total -= 1
        bits = [b for b in free_bits]
        bits[sector] = '0'
        self.entries[track] = (total, ''.join(bits))



import unittest

from d64.d64_bam import D64BAM
from d64.d64_image import D64Image
from d64.exceptions import ConsistencyError

from test.mock_block import MockBlock


class TestBAM(unittest.TestCase):

    def setUp(self):
        self.dir_data = b'\x12\x01A\x00\x15\xff\xff\x1f\x15\xff\xff\x1f\x15\xff\xff\x1f\x15\xff\xff\x1f\x15' \
                        b'\xff\xff\x1f\x15\xff\xff\x1f\x15\xff\xff\x1f\x04\x80\x02\n\x00\x00\x00\x00\x00\x00' \
                        b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                        b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x0el\xfb\x07\x00\x00\x00\x00\x00\x00\x00\x00' \
                        b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01' \
                        b'\x00\x02\x00\x12\xff\xff\x03\x12\xff\xff\x03\x12\xff\xff\x03\x12\xff\xff\x03\x11\xff' \
                        b'\xff\x01\x11\xff\xff\x01\x11\xff\xff\x01\x11\xff\xff\x01\x11\xff\xff\x01GAMES TAPE\xa0' \
                        b'\xa0\xa0\xa0\xa0\xa0\xa0\xa0GT\xa02A\xa0\xa0\xa0\xa0\x00\x00\x00\x00\x00\x00\x00\x00' \
                        b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                        b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                        b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                        b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        self.image = D64Image(None)
        self.image.bam_block = MockBlock()

    def test_get_entry(self):
        self.image.bam_block.data = self.dir_data
        entry = self.image.bam.get_entry(1)
        self.assertEqual(entry, (21, '111111111111111111111000'))

    def test_is_alloc_true(self):
        self.image.bam_block.data = self.dir_data
        self.assertTrue(self.image.bam.is_allocated(8, 3))

    def test_is_alloc_false(self):
        self.image.bam_block.data = self.dir_data
        self.assertFalse(self.image.bam.is_allocated(8, 7))

    def test_total_free(self):
        self.image.bam_block.data = self.dir_data
        self.assertEqual(self.image.bam.total_free(), 309)

    def test_check(self):
        self.image.bam_block.data = self.dir_data
        self.image.bam.check()

    def test_set_entry(self):
        self.image.bam_block.data = bytearray(self.dir_data)
        track = 1
        self.image.bam.set_entry(track, 15, '110111011100111101011000')
        entry_start = self.image.bam.BAM_OFFSET+(track-1)*self.image.bam.BAM_ENTRY_SIZE
        self.assertEqual(self.image.bam_block.data[entry_start], 15)
        self.assertEqual(self.image.bam_block.data[entry_start+1:entry_start+4], b'\xbb\xf3\x1a')

    def test_set_allocated(self):
        self.image.bam_block.data = bytearray(self.dir_data)
        track = 8
        self.image.bam.set_allocated(track, 7)
        entry = self.image.bam.get_entry(track)
        self.assertEqual(entry, (3, '000000000100000001010000'))

    def test_set_allocated_invalid(self):
        self.image.bam_block.data = bytearray(self.dir_data)
        with self.assertRaises(ValueError):
            self.image.bam.set_allocated(8, 3)

    def test_set_free(self):
        self.image.bam_block.data = bytearray(self.dir_data)
        track = 8
        self.image.bam.set_free(track, 2)
        entry = self.image.bam.get_entry(track)
        self.assertEqual(entry, (5, '001000010100000001010000'))

    def test_set_free_invalid(self):
        self.image.bam_block.data = bytearray(self.dir_data)
        with self.assertRaises(ValueError):
            self.image.bam.set_free(8, 7)

    def test_check_bad(self):
        self.image.bam_block.data = bytearray(self.dir_data)
        self.image.bam.set_entry(1, 14, '110111011100111101011000')
        with self.assertRaises(ConsistencyError):
            self.image.bam.check()


class TestBAMStatic(unittest.TestCase):

    def test_free_from_zero(self):
        free_bits = '001000010100000001010000'
        self.assertEqual(D64BAM.free_from(free_bits, 0), 2)

    def test_free_from_wrap(self):
        free_bits = '001000010100000000000000'
        self.assertEqual(D64BAM.free_from(free_bits, 11), 2)

    def test_free_from_bad(self):
        free_bits = '000000000000000000000000'
        with self.assertRaises(ConsistencyError):
            D64BAM.free_from(free_bits, 0)


if __name__ == '__main__':
    unittest.main()



import unittest
from unittest.mock import Mock

from d64.d80_d82_bam import D80D82BAM

from test.mock_block import MockBlock


class TestD80BAM(unittest.TestCase):

    def setUp(self):
        self.image = Mock()
        self.image.MIN_TRACK = 1
        self.image.MAX_TRACK = 77
        bam1 = MockBlock()
        bam1.track_range = (1, 51)
        bam2 = MockBlock()
        bam2.track_range = (51, 78)
        self.image.bam_blocks = [bam1, bam2]
        self.image.bam = D80D82BAM(self.image)

    def test_get_entry(self):
        self.image.bam_blocks[0].data[0x10:0x15] = b'\x12\xc9\xa7\x6e\x1b'
        entry = self.image.bam.get_entry(3)
        self.assertEqual(entry, (18, '10010011111001010111011011011000'))
        self.image.bam_blocks[1].data[0x10:0x15] = b'\x0f\x95\x7c\x30\x3a'
        entry = self.image.bam.get_entry(53)
        self.assertEqual(entry, (15, '10101001001111100000110001011100'))

    def test_set_entry(self):
        self.image.bam.set_entry(3, 15, '10101001001111100000110001011100')
        self.assertEqual(self.image.bam_blocks[0].data[0x10:0x15], b'\x0f\x95\x7c\x30\x3a')
        self.image.bam.set_entry(53, 18, '10010011111001010111011011011000')
        self.assertEqual(self.image.bam_blocks[1].data[0x10:0x15], b'\x12\xc9\xa7\x6e\x1b')


if __name__ == '__main__':
    unittest.main()
